from typing import TYPE_CHECKING, Type

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.geography.serializers import CountryRepresentationSerializer
from wbcore.contrib.tags.serializers import TagRepresentationSerializer

from wbfdm.models import Instrument, InstrumentType

from ..exchanges import ExchangeRepresentationSerializer
from .classifications import ClassificationRepresentationSerializer
from .mixins import InstrumentAdditionalResourcesMixin

if TYPE_CHECKING:
    User = Type[get_user_model()]


class InstrumentTypeRepresentationSerializer(serializers.RepresentationSerializer):
    class Meta:
        model = InstrumentType
        fields = ("id", "name", "key")


class InstrumentRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbfdm:instrument-detail")

    def get_filter_params(self, request):
        filter_params = {}
        if (view := request.parser_context.get("view", None)) and (
            classification_id := view.kwargs.get("classification_id", None)
        ):
            filter_params["classifications_neq"] = classification_id

        return filter_params

    class Meta:
        model = Instrument
        fields = ("id", "name", "ticker", "isin", "computed_str", "_detail")


class ClassifiableInstrumentRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["instrument_type__is_classifiable"] = True
        filter_params["level"] = 0
        return filter_params


class SecurityRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["is_security"] = True
        return filter_params


class InvestableUniverseRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["is_investable_universe"] = True
        return filter_params


class InvestableInstrumentRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["is_investable"] = True
        return filter_params


class PrimaryInvestableInstrumentRepresentationSerializer(InvestableInstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["is_primary"] = True
        return filter_params


class ManagedInstrumentRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["is_managed"] = True
        return filter_params


class EquityRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["investment_type__key"] = "equity"
        return filter_params


class ProductRepresentationSerializer(InstrumentRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = super().get_filter_params(request)
        filter_params["investment_type__key"] = "product"
        return filter_params


class InstrumentModelListSerializer(
    serializers.ModelSerializer,
):
    _currency = CurrencyRepresentationSerializer(source="currency")
    currency = serializers.PrimaryKeyRelatedField(queryset=Currency.objects.all())
    currency_symbol = serializers.CharField(read_only=True)
    _exchange = ExchangeRepresentationSerializer(source="exchange")
    country = serializers.PrimaryKeyRelatedField(queryset=Geography.countries.all())
    _country = CountryRepresentationSerializer(source="country")
    instrument_type = serializers.PrimaryKeyRelatedField(queryset=InstrumentType.objects.all())
    _instrument_type = InstrumentTypeRepresentationSerializer(source="instrument_type")
    _tags = TagRepresentationSerializer(many=True, source="tags")

    _classifications = ClassificationRepresentationSerializer(
        source="classifications", many=True, label_key="{{ name }}"
    )
    _parent = InstrumentRepresentationSerializer(source="parent")
    is_active = serializers.BooleanField(default=True, read_only=True)
    is_cash = serializers.BooleanField(default=False, read_only=True)
    _group_key = serializers.CharField(read_only=True)

    class Meta:
        model = Instrument
        read_only_fields = (
            "name",
            "exchange",
            "computed_str",
            "instrument_type",
            "old_isins",
            "inception_date",
            "country",
            "is_active",
            "parent",
            "is_primary",
            "is_investable_universe",
            "is_security",
            "is_managed",
        )

        fields = (
            "id",
            "name",
            "name_repr",
            "description",
            "instrument_type",
            "_instrument_type",
            "inception_date",
            "delisted_date",
            "country",
            "_country",
            "currency",
            "_currency",
            "currency_symbol",
            "isin",
            "refinitiv_identifier_code",
            "refinitiv_mnemonic_code",
            "ticker",
            "_classifications",
            "classifications",
            "_tags",
            "tags",
            "is_active",
            "is_cash",
            "_parent",
            "parent",
            "_additional_resources",
            "_buttons",
            "_group_key",
            "exchange",
            "_exchange",
            "is_primary",
            "is_investable_universe",
            "is_security",
            "is_managed",
        )


class InstrumentModelSerializer(InstrumentAdditionalResourcesMixin, InstrumentModelListSerializer):
    _related_instruments = SecurityRepresentationSerializer(source="related_instruments", many=True)

    @serializers.register_resource()
    def load_resources(self, instance, request, user, **kwargs):
        res = {
            "prices": reverse("wbfdm:prices-list", args=[instance.id], request=request),
            "financial-statistics": reverse(
                "wbfdm:instrument-financialstatistics-list", args=[instance.id], request=request
            ),
        }
        if not instance.is_managed:
            res.update(
                {
                    "fin-summary": reverse("wbfdm:financial-summary-list", args=[instance.id], request=request),
                    "swe-income-statement": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "income"], request=request
                    ),
                    "swe-balance-sheet": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "balancesheet"], request=request
                    ),
                    "swe-cashflow-statement": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "cashflow"], request=request
                    ),
                    "swe-ratios": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "ratios"], request=request
                    ),
                    "swe-margins": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "margins"], request=request
                    ),
                    "swe-summary": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "summary"], request=request
                    ),
                    "swe-cashflow-ratios": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "cashflow-ratios"], request=request
                    ),
                    "swe-asset-turnover-ratios": reverse(
                        "wbfdm:statementwithestimates-list",
                        args=[instance.id, "asset-turnover-ratios"],
                        request=request,
                    ),
                    "swe-credit": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "credit"], request=request
                    ),
                    "swe-long-term-solvency": reverse(
                        "wbfdm:statementwithestimates-list", args=[instance.id, "long-term-solvency"], request=request
                    ),
                    "swe-short-term-liquidity": reverse(
                        "wbfdm:statementwithestimates-list",
                        args=[instance.id, "short-term-liquidity"],
                        request=request,
                    ),
                    "valuation_ratios-new": reverse(
                        "wbfdm:valuation_ratios-list", args=[instance.id], request=request
                    ),
                    "income-statement": reverse(
                        "wbfdm:statement-list", args=[instance.id, "income-statement"], request=request
                    ),
                    "balance-sheet": reverse(
                        "wbfdm:statement-list", args=[instance.id, "balance-sheet"], request=request
                    ),
                    "cash-flow-statement": reverse(
                        "wbfdm:statement-list", args=[instance.id, "cash-flow-statement"], request=request
                    ),
                    "officers": reverse(viewname="wbfdm:officers-list", args=[instance.id], request=request),
                    "controversies": reverse(viewname="wbfdm:controversies-list", args=[instance.id], request=request),
                    "pai": reverse(viewname="wbfdm:pai-list", args=[instance.id], request=request),
                }
            )
        return res

    @serializers.register_only_instance_resource()
    def instance_resource(self, instance, request, user, **kwargs):
        content_type = ContentType.objects.get_for_model(instance)
        return {
            "children": reverse("wbfdm:instrument-children-list", args=[instance.id], request=request),
            "news": f'{reverse("wbnews:newsrelationship-list", args=[], request=request)}?content_type={content_type.id}&object_id={instance.id}',
        }

    @serializers.register_resource()
    def register_market_data(self, instance, request, user, **kwargs):
        return {
            "market_data": reverse("wbfdm:market_data-list", args=[instance.id], request=request),
            "cumulativereturn": f'{reverse("wbfdm:market_data-list", args=[instance.id], request=request)}?chart_type=ret',
            "drawdown": f'{reverse("wbfdm:market_data-list", args=[instance.id], request=request)}?chart_type=drawdown',
            "performance_summary": reverse("wbfdm:performance_summary-list", args=[instance.id], request=request),
            "monthly_performances": reverse("wbfdm:monthly_performances-list", args=[instance.id], request=request),
        }

    class Meta(InstrumentModelListSerializer.Meta):
        fields = InstrumentModelListSerializer.Meta.fields + (
            "refinitiv_mnemonic_code",
            "identifier",
            "base_color",
            "old_isins",
            "is_cash",
            "related_instruments",
            "_related_instruments",
        )



from datetime import date
from unittest.mock import patch

import pandas as pd
import pytest
from faker import Faker

from wbfdm.analysis.financial_analysis.statement_with_estimates import (
    StatementWithEstimates,
)
from wbfdm.analysis.financial_analysis.utils import Loader
from wbfdm.enums import Financial, MarketData

fake = Faker()


@pytest.mark.django_db
class TestStatementWithEstimates:
    # Coverage test to cover all the Statement with estimates properties
    @pytest.fixture()
    def df(self, values):
        data = []
        for financial in values:
            for interim in ["Y", "S1", "S2"]:
                data.append(
                    {
                        "year": 2021,
                        "estimate": False,
                        "interim": interim,
                        "period_end_date": date(2021, 12, 31),
                        "financial": financial.value,
                        "value": fake.pyfloat(),
                    }
                )
        df = pd.DataFrame(data).pivot_table(
            index=["year", "interim", "estimate", "period_end_date"], columns="financial", values="value"
        )
        return df.reset_index(level=[2, 3])

    @pytest.fixture()
    def summery_with_estimate_df(self):
        values = [
            Financial.REVENUE,
            Financial.GROSS_PROFIT,
            Financial.GROSS_PROFIT_MARGIN,
            Financial.NET_INCOME_REPORTED,
            Financial.EPS_REPORTED,
            Financial.EPS,
            Financial.EBITDA,
            Financial.EBIT,
            Financial.CASH_FLOW_PER_SHARE,
            Financial.BOOK_VALUE_PER_SHARE,
            Financial.CAPEX,
            Financial.SHARES_OUTSTANDING,
            Financial.TOTAL_ASSETS,
            Financial.CURRENT_LIABILITIES,
            Financial.ENTERPRISE_VALUE,
            Financial.CASH_EQUIVALENTS,
            MarketData.CLOSE,
            MarketData.MARKET_CAPITALIZATION,
            Financial.EMPLOYEES,
            Financial.CASH_AND_SHORT_TERM_INVESTMENT,
            Financial.DILUTED_WEIGHTED_AVG_SHARES,
            Financial.TOTAL_DEBT,
            Financial.NET_DEBT,
            Financial.STOCK_COMPENSATION,
            Financial.TANGIBLE_BOOK_VALUE_PER_SHARE,
        ]

        data = []
        for financial in values:
            for interim in ["Y", "S1", "S2"]:
                data.append(
                    {
                        "year": 2021,
                        "estimate": False,
                        "interim": interim,
                        "period_end_date": date(2021, 12, 31),
                        "financial": financial.value,
                        "value": fake.pyfloat(),
                    }
                )
        df = pd.DataFrame(data).pivot_table(
            index=["year", "interim", "estimate", "period_end_date"], columns="financial", values="value"
        )
        return df.reset_index(level=[2, 3])

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.NET_DEBT,
                Financial.EBITDA,
                Financial.TOTAL_DEBT,
                Financial.FREE_CASH_FLOW,
                Financial.CASH_FLOW_FROM_OPERATIONS,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_credit_with_estimates(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).credit_with_estimates
        assert set(res.df.columns) == {
            "net_debt_ebitda_ratio",
            "total_debt_ebitda_ratio",
            "fcf_total_debt_ratio",
            "cfo_total_debt_ratio",
            "period_end_date",
            "estimate",
        }
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.TOTAL_DEBT,
                Financial.SHAREHOLDERS_EQUITY,
                Financial.WORKING_CAPITAL,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_long_term_solvency_with_estimates(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).long_term_solvency_with_estimates
        assert set(res.df.columns) == {
            "total_debt_equity_ratio",
            "total_debt_capital_ratio",
            "period_end_date",
            "estimate",
        }
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.CURRENT_ASSETS,
                Financial.CURRENT_LIABILITIES,
                Financial.CASH_FLOW_FROM_OPERATIONS,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_short_term_liquidity_with_estimates(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).short_term_liquidity_with_estimates
        assert set(res.df.columns) == {
            "current_ratio",
            "cash_from_operation_current_liabilities_ratio",
            "period_end_date",
            "estimate",
        }
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.REVENUE,
                Financial.TOTAL_ASSETS,
                Financial.COST_OF_GOODS_SOLD,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_asset_turnover_with_estimates(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).asset_turnover_with_estimates
        assert set(res.df.columns) == {"asset_turnover", "period_end_date", "estimate"}
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.CASH_FLOW_FROM_OPERATIONS,
                Financial.CURRENT_LIABILITIES,
                Financial.REVENUE,
                Financial.TOTAL_ASSETS,
                Financial.TOTAL_DEBT,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_cashflow_ratios_with_estimates(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).cashflow_ratios_with_estimates
        assert set(res.df.columns) == {
            "current_liability_coverage_ratio",
            "cash_flow_margin",
            "asset_efficiency_margin",
            "net_debt_coverage_ratio",
            "period_end_date",
            "estimate",
        }
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.REVENUE,
                Financial.GROSS_PROFIT,
                Financial.SGA_EXPENSES,
                Financial.EBITDA,
                Financial.EBIT,
                Financial.NET_INCOME,
                Financial.FREE_CASH_FLOW,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_margins_with_estimates(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).margins_with_estimates
        assert set(res.df.columns) == {
            "gross_margin",
            "sga_margin",
            "ebitda_margin",
            "ebit_margin",
            "net_income_margin",
            "free_cash_flow_margin",
            "period_end_date",
            "estimate",
        }
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.REVENUE,
                Financial.COST_OF_GOODS_SOLD,
                Financial.GROSS_PROFIT,
                Financial.GROSS_PROFIT_MARGIN,
                Financial.SELLING_MARKETING_EXPENSES,
                Financial.SGA_EXPENSES,
                Financial.GENERAL_ADMIN_EXPENSES,
                Financial.RND_EXPENSES,
                Financial.STOCK_COMPENSATION,
                Financial.TOTAL_OPERATING_EXPENSES,
                Financial.EBITDA,
                Financial.EBITDA_PER_SHARE,
                Financial.DEPRECATION,
                Financial.AMORTIZATION,
                Financial.EBIT,
                Financial.INTEREST_EXPENSE,
                Financial.NET_INCOME_BEFORE_TAXES,
                Financial.TAX_PROVISION,
                Financial.TAX_RATE,
                Financial.NET_INCOME,
                Financial.SHARES_OUTSTANDING,
                Financial.EPS,
                Financial.EBITDA_REPORTED,
                Financial.NET_INCOME_BEFORE_TAXES_REPORTED,
                Financial.NET_INCOME_REPORTED,
                Financial.EPS_REPORTED,
                Financial.DIVIDEND_PER_SHARE,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_income_statement_with_estimate(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).income_statement_with_estimate
        assert set(res.df.columns) == {*[v.value for v in values], "period_end_date", "estimate"}
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.CASH_EQUIVALENTS,
                Financial.INVENTORY,
                Financial.CURRENT_ASSETS,
                Financial.TOTAL_ASSETS,
                Financial.CURRENT_LIABILITIES,
                Financial.CURRENT_DEFERRED_REVENUE,
                Financial.TOTAL_DEBT,
                Financial.NET_DEBT,
                Financial.SHAREHOLDERS_EQUITY,
                Financial.GOODWILL,
                Financial.NET_ASSET_VALUE,
                Financial.BOOK_VALUE_PER_SHARE,
                Financial.TANGIBLE_BOOK_VALUE_PER_SHARE,
                Financial.ENTERPRISE_VALUE,
                Financial.TANGIBLE_BOOK_VALUE,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_balance_sheet_with_estimate(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).balance_sheet_with_estimate
        assert set(res.df.columns) == {*[v.value for v in values], "period_end_date", "estimate"}
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.WORKING_CAPITAL,
                Financial.INCOME_TAXES_PAID,
                Financial.CASH_FLOW_FROM_OPERATIONS,
                Financial.CAPEX,
                Financial.CASH_FLOW_FROM_INVESTING,
                Financial.FREE_CASH_FLOW,
                Financial.FREE_CASH_FLOW_PER_SHARE,
                Financial.TOTAL_DIVIDENDS,
                Financial.CASH_FLOW_FROM_FINANCING,
                Financial.CASH_FLOW_PER_SHARE,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_cash_flow_statement_with_estimate(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).cash_flow_statement_with_estimate
        assert set(res.df.columns) == {*[v.value for v in values], "period_end_date", "estimate"}
        assert not res.formatted_df.empty

    @pytest.mark.parametrize(
        "values",
        [
            [
                Financial.RETURN_ON_EQUITY,
                Financial.RETURN_ON_INVESTED_CAPITAL,
                Financial.RETURN_ON_CAPITAL,
                Financial.RETURN_ON_ASSETS,
            ]
        ],
    )
    @patch.object(Loader, "load")
    def test_ratios_with_estimate(self, mock_fct, instrument, df, values):
        mock_fct.return_value = df
        res = StatementWithEstimates(instrument).ratios_with_estimate
        assert set(res.df.columns) == {*[v.value for v in values], "period_end_date", "estimate"}
        assert not res.formatted_df.empty

    @patch.object(Loader, "load")
    def test_summary_with_estimate(self, mock_fct, summery_with_estimate_df, instrument):
        mock_fct.return_value = summery_with_estimate_df
        res = StatementWithEstimates(instrument).summary_with_estimate
        expected_values = {
            "revenue",
            "price_sales_ratio",
            "gross_profit",
            "gross_profit_margin",
            "net_income_reported",
            "net_income_margin",
            "eps_reported",
            "price_earning_reported_ratio",
            "eps",
            "price_earning_ratio",
            "ebitda",
            "ebitda_margin",
            "ev_ebitda_ratio",
            "ebit",
            "ebit_margin",
            "ev_ebit_ratio",
            "total_assets",
            "current_liabilities",
            "roce",
            "cash_flow_per_share",
            "price_to_cash_flow_ratio",
            "book_value_per_share",
            "price_to_book_ratio",
            "tangible_book_value_per_share",
            "price_to_tangible_bv_ratio",
            "cash_and_short_term_investment",
            "diluded_weighted_avg_shares",
            "cash_shares_ratio",
            "total_debt",
            "total_debt_shares_ratio",
            "net_debt",
            "net_debt_shares_ratio",
            "stock_compensation",
            "employees",
            "stock_compensation_employee_ratio",
            "capex",
            "shares_outstanding",
            "market_capitalization",
            "close",
            "market_capitalization",
            "net_cash",
            "period_end_date",
            "estimate",
        }
        assert set(res.df.columns) == expected_values
        assert not res.formatted_df.empty



from wbcore.metadata.configs.titles import TitleViewConfig

from wbfdm.models import Instrument


class InstrumentPriceTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Instrument Prices"

    def get_instance_title(self):
        return "Price: {{_instrument.name}} {{date}}"

    def get_create_title(self):
        return "New Price"


class InstrumentPriceInstrumentTitleConfig(InstrumentPriceTitleConfig):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Prices"


class InstrumentPriceStatisticsInstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Statistics Chart"


class MonthlyPerformancesInstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Monthly Returns Table"


class InstrumentTitleConfigMixin(TitleViewConfig):
    def get_list_title(self):
        assert self.message, "No message has been set"
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{self.message} {str(instrument)}"


class FinancialStatisticsInstrumentTitleConfig(InstrumentTitleConfigMixin):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Financial Statistics Chart"


class InstrumentPriceInstrumentDistributionReturnsChartTitleConfig(InstrumentTitleConfigMixin):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Distribution Returns Chart"


class BestAndWorstReturnsInstrumentTitleConfig(InstrumentTitleConfigMixin):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Best and Worst returns Table "



from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers


class InstrumentAdditionalResourcesMixin:
    @wb_serializers.register_only_instance_resource()
    def instrument_resources(self, instance, request, user, **kwargs):
        additional_resources = dict()

        if instance.prices.exists():
            additional_resources["instrumentpricestatisticchart"] = reverse(
                "wbfdm:instrument-pricestatisticchart-list",
                args=[instance.id],
                request=request,
            )
            additional_resources["instrumentprices"] = reverse(
                "wbfdm:instrument-price-list",
                args=[instance.id],
                request=request,
            )

        additional_resources["distributionreturnschart"] = reverse(
            "wbfdm:instrument-distributionreturnschart-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["bestandworstreturns"] = reverse(
            "wbfdm:instrument-bestandworstreturns-list",
            args=[instance.id],
            request=request,
        )
        additional_resources[
            "price_and_volume"
        ] = f'{reverse("wbfdm:market_data-list", args=[instance.id], request=request)}?chart_type=close&indicators=sma_50,sma_100&volume=true'

        additional_resources["classifications_list"] = reverse(
            "wbfdm:instrument-classification-list",
            args=[instance.get_root().id],
            request=request,
        )
        additional_resources["instrument_lists"] = reverse(
            "wbfdm:instrument-instrumentlistthrough-list", args=[instance.id], request=request
        )

        return additional_resources

    @wb_serializers.register_only_instance_resource()
    def related_instruments(self, instance, request, user, **kwargs):
        return {
            "related_instruments": reverse(
                "wbfdm:instrument-relatedinstrument-list", args=[instance.id], request=request
            )
        }

    # @wb_serializers.register_only_instance_resource()
    # def deprecated_financial_resources(self, instance, request, user, **kwargs):
    #     additional_resources = dict()
    #     additional_resources["summary_table"] = reverse(
    #         "wbfdm:instrument-summarytablechart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["financials_graph"] = reverse(
    #         "wbfdm:instrument-financialsgraphchart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["profitability_ratios"] = reverse(
    #         "wbfdm:instrument-profitabilityratioschart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["cash_flow_analysis_table"] = reverse(
    #         "wbfdm:instrument-cashflowanalysistablechart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["cash_flow_analysis_chart"] = reverse(
    #         "wbfdm:instrument-cashflowanalysisbarchart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["net_debt_and_ebitda_chart"] = reverse(
    #         "wbfdm:instrument-netdebtandebitdachart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     earnings_base_url = reverse(
    #         "wbfdm:instrument-earningschart-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["earnings_chart"] = earnings_base_url
    #     additional_resources["earnings_chart_ttm"] = f"{earnings_base_url}?period=TTM"
    #     additional_resources["earnings_chart_ntm"] = f"{earnings_base_url}?period=FTM"
    #     valuation_ratios_base_url = reverse(
    #         "wbfdm:instrument-valuationratios-list",
    #         args=[instance.id],
    #         request=request,
    #     )
    #     additional_resources["valuation_ratios-old"] = valuation_ratios_base_url
    #     additional_resources["valuation_ratios_ranges"] = f"{valuation_ratios_base_url}?ranges=true"
    #     additional_resources["valuation_ratios_related"] = f"{valuation_ratios_base_url}?vs_related=true"
    #     return additional_resources



from typing import Any, Self

from django.db import models

from wbfdm.enums import (
    ESGControveryFlag,
    ESGControverySeverity,
    ESGControveryStatus,
    ESGControveryType,
)
from wbfdm.models.instruments.instruments import Instrument


class Controversy(models.Model):
    external_id = models.CharField(max_length=512, unique=True)
    instrument = models.ForeignKey("wbfdm.Instrument", on_delete=models.CASCADE, limit_choices_to=models.Q(level=0))
    headline = models.TextField(verbose_name="Headline")
    description = models.TextField(verbose_name="Description")
    source = models.TextField(
        verbose_name="Source", null=True, blank=True
    )  # Source is usually a sentence but can be a whole text
    status = models.CharField(
        max_length=64,
        choices=ESGControveryStatus.choices,
        verbose_name="Status",
        default=ESGControveryStatus.ONGOING.value,
        null=True,
        blank=True,
    )
    type = models.CharField(
        max_length=64,
        choices=ESGControveryType.choices,
        verbose_name="Type",
        default=ESGControveryType.STRUCTURAL.value,
        null=True,
        blank=True,
    )
    severity = models.CharField(
        max_length=64,
        choices=ESGControverySeverity.choices,
        verbose_name="Severity",
        default=ESGControverySeverity.MINOR.value,
        null=True,
        blank=True,
    )
    flag = models.CharField(
        max_length=64,
        choices=ESGControveryFlag.choices,
        verbose_name="Flag",
        default=ESGControveryFlag.GREEN.value,
        null=True,
        blank=True,
    )
    direct_involvement = models.BooleanField(default=True, verbose_name="Direct Involvement")
    company_response = models.CharField(max_length=512, null=True, blank=True)

    review = models.DateField(verbose_name="Reviewed")
    initiated = models.DateField(verbose_name="initiated", null=True, blank=True)

    @classmethod
    def sync_from_dataloader(cls, controversy: dict[str, Any]) -> Self:
        instrument = Instrument.objects.get(
            pk=controversy["instrument_id"]
        ).get_root()  # we link it only to root level

        return cls.objects.update_or_create(
            external_id=controversy["id"],
            defaults={
                "instrument": instrument,
                "headline": controversy["headline"],
                "description": controversy["narrative"],
                "source": controversy["source"],
                "direct_involvement": controversy.get("direct_involvement", True),
                "company_response": controversy["response"],
                "review": controversy["review"],
                "initiated": controversy["initiated"],
                "flag": controversy["flag"],
                "status": controversy["status"],
                "type": controversy["type"],
                "severity": controversy["assessment"],
            },
        )[0]



from datetime import date
from typing import Iterator, Protocol

from wbfdm.dataloaders.types import (
    AdjustmentDataDict,
    CorporateActionDataDict,
    ESGControversyDataDict,
    ESGDataDict,
    FinancialDataDict,
    MarketDataDict,
    OfficerDataDict,
    ReportDateDataDict,
    StatementDataDict,
)
from wbfdm.enums import (
    ESG,
    CalendarType,
    DataType,
    EstimateType,
    Financial,
    Frequency,
    MarketData,
    PeriodType,
    SeriesType,
    StatementType,
)


class ReportDateProtocol(Protocol):
    def reporting_dates(self, only_next: bool = True) -> Iterator[ReportDateDataDict]: ...


class AdjustmentsProtocol(Protocol):
    def adjustments(
        self, from_date: date | None = None, to_date: date | None = None
    ) -> Iterator[AdjustmentDataDict]: ...


class MarketDataProtocol(Protocol):
    def market_data(
        self,
        values: list[MarketData] = [MarketData.CLOSE],
        from_date: date | None = None,
        to_date: date | None = None,
        exact_date: date | None = None,
        frequency: Frequency = Frequency.DAILY,
        target_currency: str | None = None,
    ) -> Iterator[MarketDataDict]: ...


class CorporateActionsProtocol(Protocol):
    def corporate_actions(
        self, from_date: date | None = None, to_date: date | None = None
    ) -> Iterator[CorporateActionDataDict]: ...


class OfficersProtocol(Protocol):
    def officers(self) -> Iterator[OfficerDataDict]: ...


class StatementsProtocol(Protocol):
    def statements(
        self,
        statement_type: StatementType | None = None,
        from_date: date | None = None,
        to_date: date | None = None,
        from_year: int | None = None,
        to_year: int | None = None,
        period_type: PeriodType = PeriodType.ALL,
        data_type: DataType = DataType.STANDARDIZED,
        financials: list[Financial] | None = None,
        target_currency: str | None = None,
    ) -> Iterator[StatementDataDict]: ...


class FinancialsProtocol(Protocol):
    def financials(
        self,
        values: list[Financial],
        from_date: date | None = None,
        to_date: date | None = None,
        from_year: int | None = None,
        to_year: int | None = None,
        from_index: int | None = None,
        to_index: int | None = None,
        from_valid: date | None = None,
        to_valid: date | None = None,
        period_type: PeriodType = PeriodType.ANNUAL,
        calendar_type: CalendarType = CalendarType.FISCAL,
        series_type: SeriesType = SeriesType.COMPLETE,
        data_type: DataType = DataType.STANDARDIZED,
        estimate_type: EstimateType = EstimateType.VALID,
        target_currency: str | None = None,
    ) -> Iterator[FinancialDataDict]: ...


class ESGControversyProtocol(Protocol):
    def esg_controversies(self) -> Iterator[ESGControversyDataDict]: ...


class ESGProtocol(Protocol):
    def esg(
        self,
        values: list[ESG],
    ) -> Iterator[ESGDataDict]: ...



from django.db.models.signals import ModelSignal

# this signal is triggered by the investable universe manager in order to gather what instrument are considered within the investable universe.
add_instrument_to_investable_universe = ModelSignal(use_caching=False)

# this signal is triggered whenever prices are stored in the system and action needs to be considered
instrument_price_imported = ModelSignal(use_caching=False)

# this signal is triggered whenever all prices are imported in the system and action needs to be considered
investable_universe_updated = ModelSignal(use_caching=False)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class SummaryTableInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-summarytablechart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class FinancialsGraphInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-financialsgraphchart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class ProfitabilityRatiosInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-profitabilityratioschart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class CashFlowAnalysisInstrumentTableChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-cashflowanalysistablechart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class CashFlowAnalysisInstrumentBarChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-cashflowanalysisbarchart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class NetDebtAndEbitdaInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-netdebtandebitdachart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class EarningsInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-earningschart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )



from django.utils.translation import gettext_lazy as _
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class AbstractClassificationButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(key="treechart", label=_("Tree Chart"), icon=WBIcon.CHART_PIE.icon),
            bt.WidgetButton(key="iciclechart", label=_("Icicle Chart"), icon=WBIcon.CHART_PYRAMID.icon),
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()


class ClassificationGroupButtonConfig(AbstractClassificationButtonConfig):
    pass


class ClassificationButtonConfig(AbstractClassificationButtonConfig):
    pass



from datetime import date
from decimal import Decimal

import pandas as pd
import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from wbcore.models import DynamicDecimalField, DynamicFloatField

from wbfdm.models import Instrument, InstrumentPrice, RelatedInstrumentThroughModel

fake = Faker()


@pytest.mark.django_db
class TestInstrumentPriceModel:
    def test_init(self, instrument_price):
        assert instrument_price.id is not None

    def test_str(self, instrument_price):
        assert (
            str(instrument_price)
            == f"{instrument_price.instrument.name}: {instrument_price.net_value} {instrument_price.date:%d.%m.%Y}"
        )

    def test_previous_price(self, instrument, instrument_price_factory):
        previous_price = instrument_price_factory.create(instrument=instrument, date=date(2009, 12, 31))
        price = instrument_price_factory.create(instrument=instrument, date=date(2010, 1, 1))
        assert price.previous_price == previous_price

    def test_previous_price_does_not_exist(self, instrument_price):
        assert instrument_price.previous_price is None

    def test_next_price_price(self, instrument, instrument_price_factory):
        next_price = instrument_price_factory.create(instrument=instrument, date=date(2010, 1, 4))
        price = instrument_price_factory.create(instrument=instrument, date=date(2010, 1, 1))
        assert price.next_price == next_price

    def test_next_price_does_not_exist(self, instrument_price):
        assert instrument_price.next_price is None

    def test_subquery_closest_value(self, instrument, instrument_price_factory):
        instrument_price_factory.create_batch(5, instrument=instrument)
        latest_valid_price = InstrumentPrice.objects.latest("date")
        queryset = Instrument.objects.all().annotate(
            net_value=InstrumentPrice.subquery_closest_value(
                "net_value", val_date=latest_valid_price.date, instrument_pk_name="pk"
            )
        )

        assert queryset.get(id=instrument.id).net_value == latest_valid_price.net_value

    def test_compute_and_update_statistics(self, weekday, instrument_factory, instrument_price_factory):
        product = instrument_factory.create()
        benchmark = instrument_factory.create()
        RelatedInstrumentThroughModel.objects.create(
            instrument=product,
            related_instrument=benchmark,
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.BENCHMARK,
        )
        risk_free_rate = instrument_factory.create()
        RelatedInstrumentThroughModel.objects.create(
            instrument=product,
            related_instrument=risk_free_rate,
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.RISK_INSTRUMENT,
        )

        for _d in pd.date_range(end=weekday, periods=5, freq="B"):
            instrument_price_factory.create(
                calculated=False, instrument=benchmark, date=_d.date(), sharpe_ratio=None, correlation=None, beta=None
            )
            instrument_price_factory.create(
                calculated=False,
                instrument=risk_free_rate,
                date=_d.date(),
                sharpe_ratio=None,
                correlation=None,
                beta=None,
            )
            instrument_price_factory.create(
                calculated=False, instrument=product, date=_d.date(), sharpe_ratio=None, correlation=None, beta=None
            )

        p = product.valuations.get(date=weekday)
        p.compute_and_update_statistics(min_period=5)
        assert p.sharpe_ratio
        assert p.correlation
        assert p.beta

    # Compute functions tests
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_180d", [None])
    # def test_compute_custom_beta_180d(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_180d")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_180d"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_1y", [None])
    # def test_compute_custom_beta_1y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_1y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_1y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_2y", [None])
    # def test_compute_custom_beta_2y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_2y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_2y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_3y", [None])
    # def test_compute_custom_beta_3y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_3y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_3y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_5y", [None])
    # def test_compute_custom_beta_5y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_5y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_5y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize(
    #     "instrument_price__free_float_ratio, instrument_price__outstanding_shares", [(None, decimal.Decimal(10))]
    # )
    # def test_compute_free_float_ratio(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_free_float_ratio")
    #     assert isinstance(instrument_price._meta.get_field("free_float_ratio"), DynamicFloatField)
    #     assert instrument_price.free_float_ratio == instrument_price.free_float / float(
    #         instrument_price.outstanding_shares
    #     )
    #
    # @pytest.mark.parametrize(
    #     "instrument_price__short_interest_ratio, instrument_price__outstanding_shares", [(None, decimal.Decimal(10))]
    # )
    # def test_compute_short_interest_ratio(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_short_interest_ratio")
    #     assert isinstance(instrument_price._meta.get_field("short_interest_ratio"), DynamicFloatField)
    #     assert instrument_price.short_interest_ratio == instrument_price.short_interest / float(
    #         instrument_price.outstanding_shares
    #     )
    #
    # @pytest.mark.parametrize("instrument_price__net_value_usd", [None])
    # def test_compute_net_value_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_net_value_usd")
    #     assert isinstance(instrument_price._meta.get_field("net_value_usd"), DynamicFloatField)
    #     assert instrument_price.net_value_usd == instrument_price.currency_fx_usd * float(instrument_price.net_value)
    #
    # @pytest.mark.parametrize("instrument_price__volume_usd", [None])
    # def test_compute_volume_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_volume_usd")
    #     assert isinstance(instrument_price._meta.get_field("volume_usd"), DynamicFloatField)
    #     assert instrument_price.volume_usd == instrument_price.currency_fx_usd * instrument_price.volume * float(
    #         instrument_price.net_value
    #     )
    #
    # @pytest.mark.parametrize("instrument_price__volume_50d_usd", [None])
    # def test_compute_volume_50d_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_volume_50d_usd")
    #     assert isinstance(instrument_price._meta.get_field("volume_50d_usd"), DynamicFloatField)
    #     assert instrument_price.volume_50d_usd == instrument_price.net_value_usd * instrument_price.volume_50d
    #
    # @pytest.mark.parametrize("instrument_price__currency_fx_usd", [None])
    # def test_compute_currency_fx_usd(self, instrument_price):
    #     fx_rate = instrument_price.instrument.currency.fx_rates.get(date=instrument_price.date)
    #     instrument_price.save(update_all_dynamic_fields=True)
    #     assert hasattr(instrument_price, "_compute_currency_fx_usd")
    #     assert isinstance(instrument_price._meta.get_field("currency_fx_usd"), DynamicFloatField)
    #     assert instrument_price.currency_fx_usd == 1 / float(fx_rate.value)

    @pytest.mark.parametrize("instrument_price__gross_value", [None])
    def test_compute_gross_value(self, instrument_price):
        assert hasattr(instrument_price, "_compute_gross_value")
        assert isinstance(instrument_price._meta.get_field("gross_value"), DynamicDecimalField)
        assert instrument_price.gross_value == instrument_price.net_value

    @pytest.mark.parametrize("instrument_price__outstanding_shares", [Decimal(10)])
    def test_compute_outstanding_shares(self, instrument_price, instrument_price_factory):
        next_price = instrument_price_factory.create(
            instrument=instrument_price.instrument,
            date=instrument_price.date + BDay(1),
            outstanding_shares=None,
            calculated=instrument_price.calculated,
        )
        assert hasattr(instrument_price, "_compute_outstanding_shares")
        assert isinstance(instrument_price._meta.get_field("outstanding_shares"), DynamicDecimalField)
        assert next_price.outstanding_shares == instrument_price.outstanding_shares

    @pytest.mark.parametrize("instrument_price__volume_50d", [None])
    def test_compute_volume_50d(self, instrument_price, instrument_price_factory):
        assert hasattr(instrument_price, "_compute_volume_50d")
        assert isinstance(instrument_price._meta.get_field("volume_50d"), DynamicFloatField)

        sum_volume_50d = instrument_price.volume
        for _d in pd.date_range(end=instrument_price.date, periods=50, freq="B", inclusive="left"):
            p = instrument_price_factory.create(date=_d, instrument=instrument_price.instrument)
            sum_volume_50d += p.volume
        instrument_price.save(update_all_dynamic_fields=True)
        assert instrument_price.volume_50d == sum_volume_50d / 50

    # @pytest.mark.parametrize("instrument_price__volume_200d", [None])
    # def test_compute_volume_200d(self, instrument_price, instrument_price_factory):
    #     assert hasattr(instrument_price, "_compute_volume_200d")
    #     assert isinstance(instrument_price._meta.get_field("volume_200d"), DynamicFloatField)
    #
    #     sum_volume_200d = instrument_price.volume
    #     for _d in pd.date_range(end=instrument_price.date, periods=200, freq="B", inclusive="left"):
    #         p = instrument_price_factory.create(date=_d, instrument=instrument_price.instrument)
    #         sum_volume_200d += p.volume
    #     instrument_price.save(update_all_dynamic_fields=True)
    #     assert instrument_price.volume_200d == sum_volume_200d / 200

    # @pytest.mark.parametrize("instrument_price__market_capitalization_usd", [None])
    # def test_compute_market_capitalization_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "market_capitalization_usd")
    #     assert isinstance(instrument_price._meta.get_field("market_capitalization_usd"), DynamicFloatField)
    #     assert (
    #         instrument_price.market_capitalization_usd
    #         == instrument_price.market_capitalization * instrument_price.currency_fx_usd
    #     )

    # def test_compute_performance_1d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(1),
    #         performance_1d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_1d")
    #     assert isinstance(instrument_price._meta.get_field("performance_1d"), DynamicDecimalField)
    #     assert instrument_price.performance_1d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_7d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(7),
    #         performance_7d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_7d")
    #     assert isinstance(instrument_price._meta.get_field("performance_7d"), DynamicDecimalField)
    #     assert instrument_price.performance_7d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_30d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(30),
    #         performance_30d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_30d")
    #     assert isinstance(instrument_price._meta.get_field("performance_30d"), DynamicDecimalField)
    #     assert instrument_price.performance_30d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_90d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(90),
    #         performance_90d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_90d")
    #     assert isinstance(instrument_price._meta.get_field("performance_90d"), DynamicDecimalField)
    #     assert instrument_price.performance_90d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_365d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(365),
    #         performance_365d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_365d")
    #     assert isinstance(instrument_price._meta.get_field("performance_365d"), DynamicDecimalField)
    #     assert instrument_price.performance_365d == instrument_price.net_value / previous_price.net_value - 1
    #
    # @pytest.mark.parametrize("instrument_price__performance_ytd", [None])
    # def test_compute_performance_ytd(self, instrument_price, instrument_price_factory):
    #     previous_price = instrument_price_factory.create(
    #         instrument=instrument_price.instrument,
    #         calculated=instrument_price.calculated,
    #         date=(date(instrument_price.date.year, 1, 1) - BDay(0)).date(),
    #         performance_ytd=None,
    #     )
    #     instrument_price.save(update_all_dynamic_fields=True)
    #     assert hasattr(instrument_price, "performance_ytd")
    #     assert isinstance(instrument_price._meta.get_field("performance_ytd"), DynamicDecimalField)
    #     assert instrument_price.performance_ytd == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_inception(self, instrument_price_factory):
    #     inception_price = instrument_price_factory.create()
    #     inception_price.instrument.inception_date = inception_price.date
    #     inception_price.instrument.save()
    #     instrument_price_factory.create(
    #         instrument=inception_price.instrument,
    #         calculated=inception_price.calculated,
    #         date=inception_price.date + BDay(1),
    #         performance_inception=None,
    #     )
    #     instrument_price = instrument_price_factory.create(
    #         instrument=inception_price.instrument,
    #         calculated=inception_price.calculated,
    #         date=inception_price.date + BDay(2),
    #         performance_inception=None,
    #     )
    #     assert hasattr(instrument_price, "performance_inception")
    #     assert isinstance(instrument_price._meta.get_field("performance_inception"), DynamicDecimalField)
    #     assert instrument_price.performance_inception == instrument_price.net_value / inception_price.net_value - 1



from datetime import timedelta

import factory
from faker import Faker
from pandas.tseries.offsets import BDay

from wbfdm.models import Option, OptionAggregate

fake = Faker()


class OptionAggregateFactory(factory.django.DjangoModelFactory):
    type = factory.Iterator(Option.Type)
    date = factory.LazyAttribute(lambda o: (fake.date_object() + BDay(1)).date())
    instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")

    volume = factory.Faker("pyfloat")
    volume_5d = factory.Faker("pyfloat")
    volume_20d = factory.Faker("pyfloat")
    volume_50d = factory.Faker("pyfloat")

    open_interest = factory.Faker("pyfloat")
    volatility_30d = factory.Faker("pyfloat")
    volatility_60d = factory.Faker("pyfloat")
    volatility_90d = factory.Faker("pyfloat")

    class Meta:
        django_get_or_create = ("type", "instrument", "date")
        model = OptionAggregate


class OptionFactory(factory.django.DjangoModelFactory):
    type = factory.Iterator(Option.Type)
    contract_identifier = factory.Faker("word")
    date = factory.LazyAttribute(lambda o: (fake.date_object() + BDay(1)).date())
    expiration_date = factory.LazyAttribute(lambda o: o.date + timedelta(days=7))
    instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    strike = factory.Faker("pyfloat")
    open = factory.Faker("pyfloat")
    high = factory.Faker("pyfloat")
    low = factory.Faker("pyfloat")
    close = factory.Faker("pyfloat")
    bid = factory.Faker("pyfloat")
    ask = factory.Faker("pyfloat")
    volume = factory.Faker("pyfloat")
    open_interest = factory.Faker("pyfloat")
    vwap = factory.Faker("pyfloat")
    volatility_30d = factory.Faker("pyfloat")
    volatility_60d = factory.Faker("pyfloat")
    volatility_90d = factory.Faker("pyfloat")
    risk_delta = factory.Faker("pyfloat")
    risk_theta = factory.Faker("pyfloat")
    risk_gamma = factory.Faker("pyfloat")
    risk_vega = factory.Faker("pyfloat")
    risk_rho = factory.Faker("pyfloat")
    risk_lambda = factory.Faker("pyfloat")
    risk_epsilon = factory.Faker("pyfloat")
    risk_vomma = factory.Faker("pyfloat")
    risk_vera = factory.Faker("pyfloat")
    risk_speed = factory.Faker("pyfloat")
    risk_zomma = factory.Faker("pyfloat")
    risk_color = factory.Faker("pyfloat")
    risk_ultima = factory.Faker("pyfloat")

    class Meta:
        django_get_or_create = ("contract_identifier", "type", "instrument", "date")
        model = Option



from contextlib import suppress
from datetime import date
from itertools import batched
from typing import Iterator

import pypika as pk
from django.db import ProgrammingError, connections
from pypika import functions as fn
from pypika.enums import SqlTypes
from pypika.terms import LiteralValue
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.contrib.qa.dataloaders.utils import SOURCE_DS2
from wbfdm.dataloaders.protocols import CorporateActionsProtocol
from wbfdm.dataloaders.types import CorporateActionDataDict


class DatastreamCorporateActionsDataloader(CorporateActionsProtocol, Dataloader):
    def corporate_actions(
        self,
        from_date: date | None = None,
        to_date: date | None = None,
    ) -> Iterator[CorporateActionDataDict]:
        lookup = {k: v for k, v in self.entities.values_list("dl_parameters__corporate_actions__parameters", "id")}

        cap_event = pk.Table("Ds2CapEvent")
        effective_date = fn.Cast(cap_event.EffectiveDate, SqlTypes.DATE)
        infocode = pk.Table("#ds2infocode")

        query = (
            pk.MSSQLQuery.select(
                cap_event.InfoCode.as_("external_identifier"),
                fn.Concat(cap_event.InfoCode, "_", effective_date).as_("id"),
                effective_date.as_("valuation_date"),
                SOURCE_DS2,
                cap_event.ActionTypeCode.as_("action_code"),
                cap_event.EventStatusCode.as_("event_code"),
                cap_event.NumOldShares.as_("old_shares"),
                cap_event.NumNewShares.as_("new_shares"),
                cap_event.ISOCurrCode.as_("currency"),
            )
            .from_(cap_event)
            .where(cap_event.InfoCode.isin([LiteralValue("select infocode from #ds2infocode")]))
            .orderby(cap_event.EffectiveDate, order=pk.Order.desc)
        )

        if from_date:
            query = query.where(cap_event.EffectiveDate >= from_date)

        if to_date:
            query = query.where(cap_event.EffectiveDate <= to_date)

        with connections["qa"].cursor() as cursor:
            # Create temporary table if it doesn't exist
            with suppress(ProgrammingError):
                cursor.execute(
                    pk.MSSQLQuery.create_table(infocode).columns(pk.Column("infocode", SqlTypes.INTEGER)).get_sql()
                )
                for batch in batched(lookup.keys(), 1000):
                    cursor.execute(f"insert into #ds2infocode values {','.join(map(lambda x: f'({x})', batch))};")

            cursor.execute(query.get_sql())
            for row in dictfetchall(cursor, CorporateActionDataDict):
                row["instrument_id"] = lookup[row["external_identifier"]]
                yield row

            # Clean up temporary table
            cursor.execute(pk.MSSQLQuery.drop_table(infocode).get_sql())



from wbcore.metadata.configs.titles import TitleViewConfig


class InstrumentESGPAITitleViewConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{self.view.instrument} - PAI"


class InstrumentESGControversiesTitleViewConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{self.view.instrument} - Controversies"



from collections import OrderedDict, defaultdict
from typing import TYPE_CHECKING, Type

from django.db.models import Model

from .dto import MetricKey

if TYPE_CHECKING:
    from wbfdm.contrib.metric.backends.base import AbstractBackend


# The cached metric backend registry as a dictionary of dictionary (e.g. {key: {model_class: {backend_class}})
class BackendRegistry:
    _key_label_map: dict[str, str] = dict()
    _metric_key_map: dict[str, MetricKey] = dict()
    _overrode_model_class: dict[MetricKey, list[Type[Model]]] = defaultdict(list)
    registry: dict[MetricKey, OrderedDict] = defaultdict(OrderedDict)

    def __getitem__(self, index):
        key = index[0]
        if isinstance(key, str):
            key = self._metric_key_map[key]
        return self.registry[key][index[1]]

    def get_choices(self, keys: list[MetricKey] | None = None) -> list[tuple[str, str]]:
        if not keys:
            keys = list(self.registry.keys())
        return [(key.key, key.label) for key in keys]

    def set(
        self,
        metric_key: MetricKey,
        model_class: Type[Model],
        backend: Type,
        move_first: bool = False,
        override_backend: bool = False,
    ):
        self._metric_key_map[metric_key.key] = metric_key
        self._key_label_map[metric_key.key] = metric_key.label
        # we ensure that the registered key and model class pair are not already registered and lock as override
        if model_class not in self._overrode_model_class[metric_key]:
            self.registry[metric_key][model_class] = backend
            if move_first:
                self.registry[metric_key].move_to_end(model_class, last=False)
        if override_backend:
            self._overrode_model_class[metric_key].append(model_class)

    def get(
        self, metric_key: MetricKey | str | None = None, model_class: Type[Model] | None = None
    ) -> list[Type["AbstractBackend"]]:
        # Initialize the backend classes list to iterate over
        registry = self.registry
        if isinstance(metric_key, str):
            metric_key = self._metric_key_map[metric_key]
        if metric_key:
            # if key is provided, we return only the backends associated with that metric key
            if metric_key not in self.registry:
                raise ValueError(f"key {metric_key.key} does not belong to a registered backend")
            registry = {metric_key: registry[metric_key]}

        if model_class:
            # if the metric needs to be computed only for a specific basket, we filter out the backend classes related to his basket class
            registry = {
                key: {_class: _dataclass}
                for key, d in registry.items()
                for _class, _dataclass in d.items()
                if _class == model_class
            }
        backends = []
        for _, ordered_dict in registry.items():
            for _, backend in ordered_dict.items():
                if backend not in backends:
                    backends.append(backend)
        return backends

    def keys(self):
        return self._metric_key_map.keys()


backend_registry = BackendRegistry()



import math
from contextlib import suppress
from datetime import date

import numpy as np
import pandas as pd
from django.db.models import (
    AutoField,
    Exists,
    ExpressionWrapper,
    F,
    OuterRef,
    QuerySet,
    Subquery,
)
from django.db.models.functions import Coalesce
from wbcore.contrib.currency.models import CurrencyFXRates

from wbfdm.enums import MarketData


class InstrumentQuerySet(QuerySet):
    def annotate_classification_for_group(
        self, classification_group, classification_height: int = 0, **kwargs
    ) -> QuerySet:
        return classification_group.annotate_queryset(
            self, classification_height, "", annotation_label="ancestor_classifications", **kwargs
        )

    def annotate_base_data(self):
        base_qs = InstrumentQuerySet(self.model, using=self._db)
        return self.annotate(
            is_investable=~Exists(base_qs.filter(parent=OuterRef("pk"))),
            root=Subquery(base_qs.filter(tree_id=OuterRef("tree_id"), level=0).values("id")[:1]),
            primary_security=ExpressionWrapper(
                Coalesce(
                    Subquery(
                        base_qs.filter(
                            parent=OuterRef("pk"),
                            is_primary=True,
                            is_security=True,
                        ).values("id")[:1]
                    ),
                    F("id"),
                ),
                output_field=AutoField(),
            ),
            primary_quote=ExpressionWrapper(
                Coalesce(
                    Subquery(
                        base_qs.filter(
                            parent=OuterRef("primary_security"),
                            is_primary=True,
                        ).values("id")[:1]
                    ),
                    F("primary_security"),
                ),
                output_field=AutoField(),
            ),
        )

    def annotate_all(self):
        return self.annotate_base_data()

    @property
    def dl(self):
        """Provides access to the dataloader proxy for the entities in the QuerySet.

        This method allows for easy retrieval of the DataloaderProxy instance
        associated with the QuerySet. It enables the utilization of dataloader
        functionalities directly from the QuerySet, facilitating data fetching and
        processing tasks.

        Returns:
            DataloaderProxy: An instance of DataloaderProxy associated with the
                entities in the QuerySet.
        """
        return self.model.dl_proxy(self)

    def get_instrument_prices_from_market_data(self, from_date: date, to_date: date):
        from wbfdm.models import InstrumentPrice

        def _dict_to_object(instrument, row):
            if (
                (price_date := row.get("date"))
                and (close := row.get("close", None))
                and int(math.log10(close)) + 1 < 10
            ):
                try:
                    try:
                        p = InstrumentPrice.objects.get(instrument=instrument, date=price_date, calculated=False)
                    except InstrumentPrice.DoesNotExist:
                        p = InstrumentPrice.objects.get(instrument=instrument, date=price_date, calculated=True)
                    p.net_value = close
                    p.gross_value = close
                    p.calculated = row["calculated"]
                    p.volume = row.get("volume", p.volume)
                    p.market_capitalization = row.get("market_capitalization", p.market_capitalization)
                    p.market_capitalization_consolidated = p.market_capitalization
                    p.set_dynamic_field(False)
                    p.id = None
                    return p
                except InstrumentPrice.DoesNotExist:
                    with suppress(CurrencyFXRates.DoesNotExist):
                        p = InstrumentPrice(
                            currency_fx_rate_to_usd=CurrencyFXRates.objects.get(
                                # we need to get the currency rate because we bulk create the object, and thus save is not called
                                date=price_date,
                                currency=instrument.currency,
                            ),
                            instrument=instrument,
                            date=price_date,
                            calculated=row["calculated"],
                            net_value=close,
                            gross_value=close,
                            volume=row.get("volume", None),
                            market_capitalization=row.get("market_capitalization", None),
                        )
                        p.set_dynamic_field(False)
                        return p

        df = pd.DataFrame(
            self.dl.market_data(
                from_date=from_date,
                to_date=to_date,
                values=[MarketData.CLOSE, MarketData.VOLUME, MarketData.MARKET_CAPITALIZATION],
            )
        )
        if not df.empty:
            df["valuation_date"] = pd.to_datetime(df["valuation_date"])
            df = df.rename(columns={"valuation_date": "date"})
            df = df.drop(
                columns=df.columns.difference(
                    ["calculated", "close", "market_capitalization", "volume", "instrument_id", "date"]
                )
            )
            df["calculated"] = False

            for instrument_id, dff in df.groupby("instrument_id", group_keys=False, as_index=False):
                dff = dff.drop(columns=["instrument_id"]).set_index("date").sort_index()
                dff = dff.reindex(pd.date_range(dff.index.min(), dff.index.max(), freq="B"))
                dff[["close", "market_capitalization"]] = dff[["close", "market_capitalization"]].astype(float).ffill()
                dff.volume = dff.volume.astype(float).fillna(0)
                dff.calculated = dff.calculated.astype(bool).fillna(
                    True
                )  # we do not ffill calculated but set the to True to mark them as "estimated"/"not real"
                dff = dff.reset_index(names="date").dropna(subset=["close"])
                dff = dff.replace([np.inf, -np.inf, np.nan], None)
                instrument = self.get(id=instrument_id)
                yield from filter(
                    lambda x: x, map(lambda row: _dict_to_object(instrument, row), dff.to_dict("records"))
                )



from django.contrib import admin

from wbfdm.models import Controversy


@admin.register(Controversy)
class ControversyModelAdmin(admin.ModelAdmin):
    search_fields = (
        "headline",
        "description",
        "source",
    )
    raw_id_fields = ["instrument"]
    list_display = (
        "initiated",
        "review",
        "headline",
        "description",
        "source",
        "status",
        "type",
        "severity",
        "company_response",
    )



# Generated by Django 5.0.12 on 2025-02-20 14:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0029_alter_instrumentprice_volume'),
    ]

    operations = [
        migrations.AlterField(
            model_name='relatedinstrumentthroughmodel',
            name='related_type',
            field=models.CharField(choices=[('BENCHMARK', 'Benchmark'), ('PEER', 'Peer'), ('RISK_INSTRUMENT', 'Risk Instrument'), ('TRACKER', 'Tracker')], default='BENCHMARK', max_length=32),
        ),
    ]



from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer

from wbfdm.models.instruments import InstrumentRequest

from .instruments import InstrumentModelSerializer, InstrumentRepresentationSerializer


class InstrumentRequestRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbfdm:instrumentrequest-detail")

    class Meta:
        model = InstrumentRequest
        fields = ("id", "status", "_detail")


class InstrumentRequestModelSerializer(wb_serializers.ModelSerializer):
    _requester = PersonRepresentationSerializer(source="requester")
    _handler = PersonRepresentationSerializer(source="handler")
    _created_instrument = InstrumentRepresentationSerializer(source="created_instrument")
    notes = wb_serializers.CharField(required=False)

    def validate(self, data):
        if (not self.instance or not self.instance.requester) and (request := self.context.get("request")):
            data["requester"] = request.user.profile
        return super().validate(data)

    class Meta:
        model = InstrumentRequest
        fields = (
            "id",
            "status",
            "requester",
            "_requester",
            "_handler",
            "handler",
            "_created_instrument",
            "created_instrument",
            "notes",
            "created",
            "_additional_resources",
        )
        flatten_fields = {
            "instrument_data": wb_serializers.JSONTableField(
                serializer_class=InstrumentModelSerializer,
                required=False,
                flatten_field_names=[
                    "name",
                    "name_repr",
                    "instrument_type",
                    "refinitiv_identifier_code",
                    "refinitiv_mnemonic_code",
                    "isin",
                    "ticker",
                    "is_cash",
                    "currency",
                    "country",
                    "tags",
                    "classifications",
                ],
            )
        }



import calendar
from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from faker import Faker
from pandas.tseries.offsets import BDay

from wbfdm.dataloaders.proxies import InstrumentDataloaderProxy
from wbfdm.factories.instrument_prices import InstrumentPrice
from wbfdm.models import Instrument, RelatedInstrumentThroughModel

fake = Faker()


@pytest.mark.django_db
class TestInstrumentModel:
    def test_init(self, instrument):
        assert instrument.id is not None

    def test_get_prices(self, instrument_factory, instrument_price_factory):
        instrument = instrument_factory.create()
        other_instrument = instrument_factory.create()
        price = instrument_price_factory.create(instrument=instrument, outstanding_shares=100)
        price.refresh_from_db()
        instrument_price_factory.create(instrument=other_instrument)  # Noise
        res = list(instrument.get_prices())
        assert len(res) == 1
        assert res[0]["valuation_date"] == price.date
        assert res[0]["open"] == float(price.net_value)
        assert res[0]["close"] == float(price.net_value)
        assert res[0]["high"] == float(price.net_value)
        assert res[0]["low"] == float(price.net_value)
        assert res[0]["volume"] == price.volume
        assert res[0]["market_capitalization"] == price.market_capitalization
        assert res[0]["outstanding_shares"] == float(price.outstanding_shares)

    def test_get_price(self, instrument_factory, instrument_price_factory):
        instrument = instrument_factory.create()
        other_instrument = instrument_factory.create()
        price = instrument_price_factory.create(instrument=instrument)
        price.refresh_from_db()
        instrument_price_factory.create(instrument=other_instrument)  # Noise
        assert instrument.get_price(price.date) == float(price.net_value)
        assert instrument.get_price((price.date + BDay(1)).date()) == float(price.net_value)
        assert instrument.get_price((price.date + BDay(2)).date()) == float(price.net_value)
        assert instrument.get_price((price.date + BDay(3)).date()) == float(price.net_value)
        with pytest.raises(ValueError):
            instrument.get_price((price.date + BDay(4)).date())  # for return the latest valid price 3 days earlier.

        # if the instrument is considered cash, we always return a value of 1
        instrument.is_cash = True
        instrument.save()
        assert instrument.get_price(price.date) == Decimal(1)

    def test_extract_daily_performance_df(self):
        tidx = pd.date_range("2016-07-01", periods=4, freq="B")
        data = np.random.randn(4)
        df = pd.Series(data, index=tidx, name="close")
        res = Instrument.extract_daily_performance_df(df)
        assert res.performance[0] == df[1] / df[0] - 1

    def test_extract_monthly_performance_df(self):
        tidx = pd.date_range("2016-07-01", periods=4, freq="ME")
        data = np.random.randn(4)
        df = pd.Series(data, index=tidx, name="close")
        df.index = pd.to_datetime(df.index)
        res = Instrument.extract_monthly_performance_df(df)
        assert res.performance[0] == df[1] / df[0] - 1

    def test_extract_inception_performance_df(self):
        tidx = pd.date_range("2016-07-01", periods=4, freq="D")
        data = np.random.randn(4)
        df = pd.Series(data, index=tidx, name="close")
        res = Instrument.extract_inception_performance_df(df)

        assert res == df[3] / df[0] - 1

    def test_extract_annual_performance_df(self):
        tidx = pd.date_range("2016-07-01", periods=4, freq="BYE")
        data = np.random.randn(4)
        df = pd.Series(data, index=tidx, name="close")
        res = Instrument.extract_annual_performance_df(df)

        assert res.performance[0] == df[1] / df[0] - 1

    def test_get_monthly_return_summary(self, instrument, instrument_price_factory):
        instrument_price_factory.create_batch(10, instrument=instrument)
        res, calculated_mask = instrument.get_monthly_return_summary()
        assert "performance" in res.columns
        assert "month" in res.columns
        assert "year" in res.columns

    def test_get_monthly_return_summary_dict(self, instrument, instrument_price_factory):
        instrument_price_factory.create_batch(10, instrument=instrument)
        res = instrument.get_monthly_return_summary_dict()
        for year in res.keys():
            for month in res[year].keys():
                assert month in calendar.month_abbr or month == "annual"
                assert "performance" in res[year][month]
                if month == "annual":
                    assert res[year][month]["performance"] is not None

    def test_get_prices_df(self, instrument, instrument_price_factory):
        price = instrument_price_factory.create(instrument=instrument)
        previous_price = instrument_price_factory.create(  # noqa
            date=price.date - timedelta(days=1), instrument=instrument
        )
        assert instrument.get_prices_df(from_date=price.date)[0] == float(price.net_value)

    @patch.object(Instrument, "get_prices_df")
    def test_build_benchmark_df(self, mock_get_prices_df, instrument, instrument_factory, instrument_price_factory):
        instrument_price_factory.create_batch(5, instrument=instrument)
        start_val = InstrumentPrice.objects.earliest("date").date
        end_val = InstrumentPrice.objects.latest("date").date
        tidx = pd.date_range(start=start_val, end=end_val, freq="D")
        mock_data = pd.Series(data=np.random.random_sample((tidx.shape[0],)), index=tidx)
        mock_get_prices_df.return_value = mock_data

        # Test without risk or benchmark assigned to instrument
        res = instrument.build_benchmark_df(end_val)
        assert res.empty

        # Assign benchmark and risk instrument
        RelatedInstrumentThroughModel.objects.create(
            instrument=instrument,
            related_instrument=instrument_factory.create(),
            is_primary=True,
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.BENCHMARK,
        )
        RelatedInstrumentThroughModel.objects.create(
            instrument=instrument,
            is_primary=True,
            related_instrument=instrument_factory.create(),
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.RISK_INSTRUMENT,
        )
        instrument.refresh_from_db()
        res = instrument.build_benchmark_df(end_val)
        assert res.rate.equals(mock_data)
        assert res.benchmark_net_value.equals(mock_data)
        assert "net_value" in res.columns

    @patch.object(InstrumentDataloaderProxy, "market_data")
    def test_save_prices_in_db(self, mock_fct, instrument, instrument_price_factory, currency_fx_rates_factory):
        # we check that the import also handle missing data with a forward fill
        from_date = date(2024, 1, 1)
        to_date = date(2024, 1, 3)
        currency_fx_rates_factory.create(currency=instrument.currency, date=from_date)
        currency_fx_rates_factory.create(currency=instrument.currency, date=from_date + timedelta(days=1))
        currency_fx_rates_factory.create(currency=instrument.currency, date=to_date)
        mock_fct.return_value = [
            {
                "valuation_date": from_date,
                "close": 1,
                "volume": 1,
                "market_capitalization": 1,
                "instrument_id": instrument.id,
            },
            {
                "valuation_date": to_date,
                "close": 2,
                "volume": 2,
                "market_capitalization": 2,
                "instrument_id": instrument.id,
            },
        ]
        instrument.import_prices(from_date, to_date + timedelta(days=1))
        assert InstrumentPrice.objects.get(
            instrument=instrument, date=from_date, calculated=False
        ).net_value == Decimal(1)
        assert InstrumentPrice.objects.get(
            instrument=instrument, date=from_date + timedelta(days=1), calculated=True
        ).net_value == Decimal(1)
        assert InstrumentPrice.objects.get(instrument=instrument, date=to_date, calculated=False).net_value == Decimal(
            2
        )

    def test_security_instrument_type(self, instrument_factory, instrument_type_factory):
        company_type = instrument_type_factory.create(is_security=False)
        security_type = instrument_type_factory.create(is_security=True)
        quote_type = instrument_type_factory.create(is_security=False)

        company = instrument_factory.create(instrument_type=company_type)
        security = instrument_factory.create(instrument_type=security_type, parent=company)
        quote = instrument_factory.create(instrument_type=quote_type, parent=security)

        assert company.security_instrument_type == company_type
        assert security.security_instrument_type == security_type
        assert quote.security_instrument_type == security_type

    def test_pre_save(self, instrument, instrument_factory):
        parent = instrument_factory.create()
        instrument.parent = parent
        instrument.save()

        instrument.name_repr = "test"
        instrument.save()
        parent.refresh_from_db()
        assert parent.name_repr == "test"

        parent.name_repr = "test2"
        parent.save()
        instrument.refresh_from_db()
        assert instrument.name_repr == "test2"



from .classifications import (
    ChildClassificationParentClassificationTitleConfig,
    ClassificationClassificationGroupTitleConfig,
    ClassificationGroupTitleConfig,
    ClassificationIcicleChartTitleConfig,
    ClassificationTitleConfig,
    ClassificationTreeChartTitleConfig,
    InstrumentClassificationThroughTitleConfig,
)
from .financials_analysis import (
    CashFlowAnalysisInstrumentBarChartTitle,
    CashFlowAnalysisInstrumentTableChartTitle,
    EarningsInstrumentChartTitle,
    FinancialsGraphInstrumentChartTitle,
    NetDebtAndEbitdaInstrumentChartTitle,
    ProfitabilityRatiosInstrumentChartTitle,
    SummaryTableInstrumentChartTitle,
    FinancialAnalysisGeneratorTitleConfig,
)
from .instrument_requests import InstrumentRequestTitleConfig
from .instruments import InstrumentTitleConfig
from .instruments_relationships import (
    ClassifiedInstrumentTitleConfig,
    InstrumentFavoriteGroupTitleConfig,
)
from .exchanges import ExchangeTitleConfig
from .instrument_prices import (
    InstrumentPriceTitleConfig,
    InstrumentPriceInstrumentTitleConfig,
    InstrumentPriceStatisticsInstrumentTitleConfig,
    MonthlyPerformancesInstrumentTitleConfig,
    InstrumentTitleConfigMixin,
    FinancialStatisticsInstrumentTitleConfig,
    InstrumentPriceInstrumentDistributionReturnsChartTitleConfig,
    BestAndWorstReturnsInstrumentTitleConfig,
)
from .market_data import (
    PerformanceSummaryChartTitleConfig,
    MarketDataChartTitleConfig,
)
from .financial_ratio_analysis import ValuationRatioChartTitleConfig
from .esg import InstrumentESGPAITitleViewConfig, InstrumentESGControversiesTitleViewConfig



from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from wbcore.contrib.io.backends import AbstractDataBackend, register

from .mixin import DataBackendMixin
from .utils import Controller

DEFAULT_MAPPING = {
    "ISOCUR": "currency__key",
    "RIC": "refinitiv_identifier_code",
    "MNEM": "refinitiv_mnemonic_code",
    "ISIN": "isin",
    "GGISO": "country",
    # "NPCUR": "currency__key",
    "BNAM": "borrower",
    "ITYP": "issuer_type",
    "EXCHB": "exchanges",
    "CTYP": "coupon_type",
    "BTYP": "bond_type",
    "WC06092": "description",
    "SEGM": "exchange",
    "NAME": "name",
    "WC05601": "ticker",
    "GDIGC": "gics_classification",
    "TR5": "trbc_classification",
    "TYPE": "instrument_type",
    "WC18272": "inception_date",
    "WC18273": "inception_date",
    "WC07015": "delisted_date",
}


@register("Instrument", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    CHUNK_SIZE = 20

    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        if obj_external_ids:
            df = self.controller.get_data(obj_external_ids, list(DEFAULT_MAPPING.keys()))
            if not df.empty:
                content_file = BytesIO()
                df.to_json(content_file, orient="records")
                file_name = f"instrument_{datetime.timestamp(execution_time)}.json"
                yield file_name, content_file



from datetime import date

from django.db import models

from wbfdm.models.instruments import Instrument


class DataBackendMixin:
    def is_object_valid(self, obj: models.Model) -> bool:
        return (
            super().is_object_valid(obj)
            and obj.is_active_at_date(date.today())
            and (obj.refinitiv_identifier_code or obj.isin or obj.refinitiv_mnemonic_code)
        )

    def get_default_queryset(self):
        privates_equities = Instrument.objects.filter(instrument_type__key="private_equity")
        return Instrument.objects.exclude(id__in=privates_equities.values("id"))

    def get_provider_id(self, obj: models.Model) -> str:
        if perm_id := self.controller.fetch_perm_id(
            instrument_ric=obj.refinitiv_identifier_code,
            instrument_isin=obj.isin,
            instrument_mnemonic=obj.refinitiv_mnemonic_code,
        ):
            return perm_id
        elif (
            obj.refinitiv_identifier_code
        ):  # We default to the RIC in case permID can't be found because some instrument don't have any
            return obj.refinitiv_identifier_code



# Generated by Django 4.2.11 on 2024-04-03 08:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0015_instrument_instrument_investible_index"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrumenttype",
            name="name_repr",
            field=models.CharField(default="name", max_length=128, verbose_name="Name (Representation)"),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.1.10 on 2023-10-04 07:55

import django.db.models.deletion
import wbfdm.models.fk_fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("geography", "0001_initial"),
        ("wbfdm", "0006_instrument_officer_backend"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="country",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to={"level": 1},
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="geography.geography",
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="currency",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="currency.currency"
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="delisted_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="description",
            field=models.TextField(default=""),
        ),
        migrations.AddField(
            model_name="instrument",
            name="inception_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="instrument_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("ADR", "American Depository Receipt (ADR)"),
                    ("CF", "Close Ended Fund"),
                    ("EQ", "Equity"),
                    ("ETF", "Exchange Traded Fund"),
                    ("ETC", "Exchange Traded Commodity"),
                    ("ETN", "Exchange Traded Note"),
                    ("WARRANT", "Warrant"),
                    ("GDR", "Global Depository Receipt (GDR)"),
                    ("GS", "Genussschein"),
                    ("TRUST", "Investment Trust"),
                    ("NVDR", "Non-voting depository receipt (NVDR)"),
                    ("PREF", "Preference Share"),
                ],
                max_length=8,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="officer_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                    "officers": wbfdm.models.fk_fields.RawFieldValue(True),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="source",
            field=models.CharField(blank=True, max_length=64, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="source_id",
            field=models.CharField(blank=True, max_length=64, null=True),
        ),
        migrations.AddField(
            model_name="instrumentlookup",
            name="esg",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="instrument",
            name="esg_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "esg": wbfdm.models.fk_fields.RawFieldValue(True),
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
    ]



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig


class InstrumentPriceDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="valuation_date", label=_("Date")),
                dp.Field(key="open", label=_("Open")),
                dp.Field(key="high", label=_("High")),
                dp.Field(key="low", label=_("Low")),
                dp.Field(key="close", label=_("Close")),
                dp.Field(key="volume", label=_("Volume")),
                dp.Field(key="outstanding_shares", label=_("Oustanding Shares")),
                dp.Field(key="market_capitalization", label=_("Market Cap.")),
            ],
        )



from wbcore import serializers


class OfficerSerializer(serializers.Serializer):
    id = serializers.PrimaryKeyCharField()
    position = serializers.CharField()
    name = serializers.CharField()
    age = serializers.IntegerField()
    sex = serializers.CharField()
    start = serializers.DateField()

    class Meta:
        fields = (
            "id",
            "position",
            "name",
            "age",
            "sex",
            "start",
        )



from datetime import date
from itertools import batched  # type: ignore
from typing import Iterator

from django.db import connections
from django.template.loader import get_template
from jinjasql import JinjaSql  # type: ignore
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.dataloaders.protocols import FinancialsProtocol
from wbfdm.dataloaders.types import FinancialDataDict
from wbfdm.enums import (
    CalendarType,
    DataType,
    EstimateType,
    Financial,
    PeriodType,
    SeriesType,
)


class IBESFinancialsDataloader(FinancialsProtocol, Dataloader):
    def financials(
        self,
        values: list[Financial],
        from_date: date | None = None,
        to_date: date | None = None,
        from_year: int | None = None,
        to_year: int | None = None,
        from_index: int | None = None,
        to_index: int | None = None,
        from_valid: date | None = None,
        to_valid: date | None = None,
        period_type: PeriodType = PeriodType.ANNUAL,
        calendar_type: CalendarType = CalendarType.FISCAL,
        series_type: SeriesType = SeriesType.COMPLETE,
        data_type: DataType = DataType.STANDARDIZED,
        estimate_type: EstimateType = EstimateType.VALID,
        target_currency: str | None = None,
    ) -> Iterator[FinancialDataDict]:
        lookup = {k: v for k, v in self.entities.values_list("dl_parameters__financials__parameters", "id")}
        for batch in batched(lookup.keys(), 1000):
            sql = ""
            if series_type == SeriesType.COMPLETE:
                sql = "qa/sql/ibes/complete.sql"
                if calendar_type == CalendarType.CALENDAR:
                    sql = "qa/sql/ibes/calendarized.sql"

            elif series_type == SeriesType.FULL_ESTIMATE:
                sql = "qa/sql/ibes/estimates.sql"

            query, bind_params = JinjaSql(param_style="format").prepare_query(
                get_template(sql, using="jinja").template,
                {
                    "instruments": batch,
                    "financials": [financial.value for financial in values],
                    "estimate_type": estimate_type.value,
                    "from_date": from_date,
                    "to_date": to_date,
                    "from_year": from_year,
                    "to_year": to_year,
                    "from_index": from_index,
                    "to_index": to_index,
                    "from_valid": from_valid,
                    "to_valid": to_valid,
                    "period_type": period_type.value,
                },
            )
            with connections["qa"].cursor() as cursor:
                cursor.execute(
                    query,
                    bind_params,
                )
                for row in dictfetchall(cursor):
                    row["instrument_id"] = lookup[row["external_identifier"]]
                    row["estimate"] = bool(row["estimate"])
                    row["value"] = row.get("value", None)
                    row["difference_pct"] = row.get("difference_pct", 0)
                    row["value_high"] = row.get("value_high", row["value"])
                    row["value_low"] = row.get("value_low", row["value"])
                    row["value_amount"] = row.get("value_amount", row["value"])
                    row["value_stdev"] = row.get("value_stdev", row["value"])
                    yield row



from datetime import date

import pandas as pd


#
#
# class FundamentalComputedMixin:
#     def _compute_revenue_growth(self) -> float:
#         if (
#             self.revenue is not None
#             and (last_fundamental := self.last_fundamental)
#             and (last_revenue := last_fundamental.revenue)
#         ):
#             return self.revenue / last_revenue - 1
#
#     def _compute_gross_profit(self) -> float:
#         if self.revenue is not None and self.cost_of_good_sold is not None:
#             return self.revenue - self.cost_of_good_sold
#
#     def _compute_gross_profit_margin(self) -> float:
#         if (
#             self.gross_profit is not None
#             and self.revenue
#             and self.deprecation_and_amortization_to_sales_ratio is not None
#         ):
#             return (self.gross_profit / self.revenue) + self.deprecation_and_amortization_to_sales_ratio
#
#     def _compute_ebitda_margin(self) -> float:
#         if self.ebitda is not None and self.revenue:
#             return self.ebitda / self.revenue
#
#     def _compute_ebit_margin(self) -> float:
#         if self.ebit is not None and self.revenue:
#             return self.ebit / self.revenue
#
#     def _compute_net_profit_margin(self) -> float:
#         if self.net_profit is not None and self.revenue:
#             return self.net_profit / self.revenue
#
#     def _compute_cost_of_good_sold(self) -> float:
#         if self.cost_of_good_sold_without_depreciation is not None and self.deprecation_and_amortization is not None:
#             return self.cost_of_good_sold_without_depreciation + self.deprecation_and_amortization
#
#     def _compute_eps_growth(self) -> float:
#         if self.eps is not None and (last_fundamental := self.last_fundamental) and (last_eps := last_fundamental.eps):
#             return self.eps / last_eps - 1
#
#     def _compute_sga_to_sales_ratio(self) -> float:
#         if self.sga is not None and self.revenue:
#             return self.sga / self.revenue
#
#     def _compute_interest_expense_to_sales_ratio(self) -> float:
#         if self.interest_expense is not None and self.revenue:
#             return self.interest_expense / self.revenue
#
#     def _compute_deprecation_and_amortization_to_sales_ratio(self) -> float:
#         if self.deprecation_and_amortization is not None and self.revenue:
#             return self.deprecation_and_amortization / self.revenue
#
#     def _compute_free_cash(self) -> float:
#         if self.cash_from_operation is not None and self.capital_expenditures is not None:
#             return self.cash_from_operation - self.capital_expenditures
#
#     def _compute_free_cash_flow(self):
#         if (
#             self.free_cash is not None
#             and (price := self.last_price)
#             and (outstanding_shares_consolidated := price.outstanding_shares_consolidated)
#         ):
#             return self.free_cash / float(outstanding_shares_consolidated)
#
#     def _compute_free_cash_flow_growth(self) -> float:
#         if (
#             self.free_cash_flow is not None
#             and (last_fundamental := self.last_fundamental)
#             and (last_free_cash_flow := last_fundamental.free_cash_flow)
#         ):
#             return self.free_cash_flow / last_free_cash_flow - 1
#
#     def _compute_free_cash_flow_to_sales_ratio(self) -> float:
#         if self.free_cash_flow is not None and self.revenue:
#             return self.free_cash_flow / self.revenue
#
#     def _compute_book_value_per_share(self) -> float:
#         if self.total_assets is not None and self.total_debt is not None:
#             if (
#                 yearly_valuations := self.instrument.valuations.exclude(outstanding_shares_consolidated=0).filter(
#                     date__year=self.year,
#                     date__lte=self.date_range.upper,
#                     outstanding_shares__isnull=False,
#                 )
#             ).exists() and (
#                 last_outstanding_shares := yearly_valuations.latest("date").outstanding_shares_consolidated
#             ):
#                 return (self.total_assets - self.total_debt) / float(last_outstanding_shares)
#
#     def _compute_net_change_in_cash(self) -> float:
#         if (
#             self.cash_from_operation is not None
#             and self.investment_cash is not None
#             and self.financing_cash is not None
#         ):
#             return self.cash_from_operation - self.investment_cash + self.financing_cash
#
#     def _compute_employee_count_growth(self) -> float:
#         if self.employee_count is not None and (previous := self.last_fundamental) and previous.employee_count:
#             return self.employee_count / previous.employee_count - 1
#
#     def _compute_capex_to_sales(self) -> float:
#         if self.capital_expenditures is not None and self.revenue:
#             return self.capital_expenditures / self.revenue
#
#     def _compute_net_debt_to_ebitda_ratio(self) -> float:
#         if self.net_debt is not None and self.ebitda:
#             return self.net_debt / self.ebitda
#
#     def _compute_burn_rate(self) -> float:
#         if self.net_change_in_cash is not None:
#             return min(self.net_change_in_cash / 12, 0)
#
#     def _compute_operating_burn_rate(self) -> float:
#         if self.cash_from_operation is not None:
#             return min(self.cash_from_operation / 12, 0)
#
#     def _compute_free_cash_burn_rate(self) -> float:
#         if self.free_cash is not None:
#             return min(self.free_cash / 12, 0)
#
#     def _compute_cash_reserve_to_burn_rate_ratio(self) -> float:
#         if self.cash_and_short_term_investments is not None and self.burn_rate:
#             return abs(self.cash_and_short_term_investments / self.burn_rate / 12)
#
#     def _compute_cash_reserve_to_operating_burn_rate_ratio(self) -> float:
#         if self.cash_and_short_term_investments is not None and self.operating_burn_rate:
#             return abs(self.cash_and_short_term_investments / self.operating_burn_rate / 12)
#
#     def _compute_cash_reserve_to_free_cash_burn_rate_ratio(self) -> float:
#         if self.cash_and_short_term_investments is not None and self.free_cash_burn_rate:
#             return abs(self.cash_and_short_term_investments / self.free_cash_burn_rate / 12)
#
#     def _compute_working_capital_to_burn_rate_ratio(self) -> float:
#         if self.working_capital is not None and self.burn_rate:
#             if self.working_capital <= 0:
#                 return None
#             return abs(self.working_capital / self.burn_rate / 12)
#
#     def _compute_working_capital_to_operating_burn_rate_ratio(self) -> float:
#         if self.working_capital is not None and self.operating_burn_rate:
#             if self.working_capital <= 0:
#                 return None
#             return abs(self.working_capital / self.operating_burn_rate / 12)
#
#     def _compute_working_capital_to_free_cash_burn_rate_ratio(self) -> float:
#         if self.working_capital is not None and self.free_cash_burn_rate:
#             if self.working_capital <= 0:
#                 return None
#             return abs(self.working_capital / self.free_cash_burn_rate / 12)
#
#     def _compute_current_ratio(self) -> float:
#         if self.current_assets is not None and self.current_liabilities:
#             return self.current_assets / self.current_liabilities
#
#     def _compute_cash_and_short_term_investments_to_current_assets_ratio(self) -> float:
#         if self.cash_and_short_term_investments is not None and self.current_assets:
#             return self.cash_and_short_term_investments / self.current_assets
#
#     def _compute_rd_to_sales_ratio(self) -> float:
#         if self.cost_research_development is not None and self.revenue:
#             return self.cost_research_development / self.revenue
#
#     def _compute_interest_coverage_ratio(self) -> float:
#         if self.ebit is not None and self.interest_expense:
#             return self.ebit / self.interest_expense
#
#     def _compute_return_on_equity(self) -> float:
#         if (
#             (previous := self.last_fundamental)
#             and self.net_profit is not None
#             and self.shareholder_equity is not None
#             and (previous_shareholder := previous.shareholder_equity) is not None
#         ):
#             avg_shareholder = (self.shareholder_equity + previous_shareholder) / 2
#             if avg_shareholder:
#                 return self.net_profit / avg_shareholder
#
#     def _compute_return_on_assets(self) -> float:
#         if (
#             (previous := self.last_fundamental)
#             and self.net_profit is not None
#             and self.total_assets is not None
#             and (previous_total_assets := previous.total_assets) is not None
#         ):
#             avg_total_assets = (self.total_assets + previous_total_assets) / 2
#             if avg_total_assets:
#                 return self.net_profit / avg_total_assets
#
#     def _compute_return_on_capital_employed(self) -> float:
#         if (
#             self.ebit is not None
#             and self.total_assets
#             and self.current_liabilities
#             and (diff := self.total_assets - self.current_liabilities)
#         ):
#             return self.ebit / diff
#
#     def _compute_return_on_invested_capital(self) -> float:
#         if (
#             self.ebit is not None
#             and self.company_tax_rate is not None
#             and self.shareholder_equity is not None
#             and self.total_liabilities is not None
#             and self.cash_and_cash_equivalents is not None
#             and (previous := self.last_fundamental)
#         ):
#             if (
#                 previous.shareholder_equity is not None
#                 and previous.total_liabilities is not None
#                 and previous.cash_and_cash_equivalents is not None
#             ):
#                 divisor = (
#                     self.shareholder_equity
#                     + self.total_liabilities
#                     - self.cash_and_cash_equivalents
#                     + previous.shareholder_equity
#                     + previous.total_liabilities
#                     - previous.cash_and_cash_equivalents
#                 ) / 2
#                 if divisor:
#                     return self.ebit * (1 - self.company_tax_rate) / divisor
#
#     def _compute_revenue_to_employee_ratio(self) -> float:
#         if self.revenue is not None and self.employee_count:
#             return self.revenue / self.employee_count
#
#
# class GeographicSegmentComputedMixin:
#     def _compute_value_growth(self) -> float:
#         if self.value is not None and (previous_segment := self.last_geographic_segment) and previous_segment.value:
#             return self.value / previous_segment.value - 1
#
#
# class ForecastComputedMixin:
#     def _compute_revenue_growth_y1(self) -> float:
#         if (fundamental := self.yearly_fundamental) and self.revenue_y1 is not None and fundamental.revenue:
#             return self.revenue_y1 / fundamental.revenue - 1
#
#     def _compute_revenue_growth_y2(self) -> float:
#         if self.revenue_y2 is not None and self.revenue_y1:
#             return self.revenue_y2 / self.revenue_y1 - 1
#
#     def _compute_revenue_growth_y3(self) -> float:
#         if self.revenue_y3 is not None and self.revenue_y2:
#             return self.revenue_y3 / self.revenue_y2 - 1
#
#     def _compute_revenue_growth_y4(self) -> float:
#         if self.revenue_y4 is not None and self.revenue_y3:
#             return self.revenue_y4 / self.revenue_y3 - 1
#
#     def _compute_revenue_growth_y5(self) -> float:
#         if self.revenue_y5 is not None and self.revenue_y4:
#             return self.revenue_y5 / self.revenue_y4 - 1
#
#     def _compute_deprecation_and_amortization_y1(self):
#         if self.ebitda_margin_y1 is not None and self.ebit_margin_y1 is not None:
#             return self.ebitda_margin_y1 - self.ebit_margin_y1
#
#     def _compute_deprecation_and_amortization_y2(self):
#         if self.ebitda_margin_y2 is not None and self.ebit_margin_y2 is not None:
#             return self.ebitda_margin_y2 - self.ebit_margin_y2
#
#     def _compute_deprecation_and_amortization_y3(self):
#         if self.ebitda_margin_y3 is not None and self.ebit_margin_y3 is not None:
#             return self.ebitda_margin_y3 - self.ebit_margin_y3
#
#     def _compute_deprecation_and_amortization_y4(self):
#         if self.ebitda_margin_y4 is not None and self.ebit_margin_y4 is not None:
#             return self.ebitda_margin_y4 - self.ebit_margin_y4
#
#     def _compute_deprecation_and_amortization_y5(self):
#         if self.ebitda_margin_y5 is not None and self.ebit_margin_y5 is not None:
#             return self.ebitda_margin_y5 - self.ebit_margin_y5
#
#     def _compute_gross_profit_margin_y1(self):
#         if (
#             self.gross_profit_margin_without_depreciation_y1 is not None
#             and self.deprecation_and_amortization_y1 is not None
#         ):
#             return self.gross_profit_margin_without_depreciation_y1 + self.deprecation_and_amortization_y1
#
#     def _compute_gross_profit_margin_y2(self):
#         if (
#             self.gross_profit_margin_without_depreciation_y2 is not None
#             and self.deprecation_and_amortization_y2 is not None
#         ):
#             return self.gross_profit_margin_without_depreciation_y2 + self.deprecation_and_amortization_y2
#
#     def _compute_gross_profit_margin_y3(self):
#         if (
#             self.gross_profit_margin_without_depreciation_y3 is not None
#             and self.deprecation_and_amortization_y3 is not None
#         ):
#             return self.gross_profit_margin_without_depreciation_y3 + self.deprecation_and_amortization_y3
#
#     def _compute_gross_profit_margin_y4(self):
#         if (
#             self.gross_profit_margin_without_depreciation_y4 is not None
#             and self.deprecation_and_amortization_y4 is not None
#         ):
#             return self.gross_profit_margin_without_depreciation_y4 + self.deprecation_and_amortization_y4
#
#     def _compute_gross_profit_margin_y5(self):
#         if (
#             self.gross_profit_margin_without_depreciation_y5 is not None
#             and self.deprecation_and_amortization_y5 is not None
#         ):
#             return self.gross_profit_margin_without_depreciation_y5 + self.deprecation_and_amortization_y5
#
#     def _compute_gross_profit_y1(self):
#         if self.gross_profit_margin_y1 is not None and self.revenue_y1 is not None:
#             return self.gross_profit_margin_y1 * self.revenue_y1
#
#     def _compute_gross_profit_y2(self):
#         if self.gross_profit_margin_y2 is not None and self.revenue_y2 is not None:
#             return self.gross_profit_margin_y2 * self.revenue_y2
#
#     def _compute_gross_profit_y3(self):
#         if self.gross_profit_margin_y3 is not None and self.revenue_y3 is not None:
#             return self.gross_profit_margin_y3 * self.revenue_y3
#
#     def _compute_gross_profit_y4(self):
#         if self.gross_profit_margin_y4 is not None and self.revenue_y4 is not None:
#             return self.gross_profit_margin_y4 * self.revenue_y4
#
#     def _compute_gross_profit_y5(self):
#         if self.gross_profit_margin_y5 is not None and self.revenue_y5 is not None:
#             return self.gross_profit_margin_y5 * self.revenue_y5
#
#     def _compute_net_profit_margin_y1(self) -> float:
#         if self.reported_net_profit_y1 is not None and self.revenue_y1:
#             return self.reported_net_profit_y1 / self.revenue_y1
#
#     def _compute_net_profit_margin_y2(self) -> float:
#         if self.reported_net_profit_y2 is not None and self.revenue_y2:
#             return self.reported_net_profit_y2 / self.revenue_y2
#
#     def _compute_net_profit_margin_y3(self) -> float:
#         if self.reported_net_profit_y3 is not None and self.revenue_y3:
#             return self.reported_net_profit_y3 / self.revenue_y3
#
#     def _compute_net_profit_margin_y4(self) -> float:
#         if self.reported_net_profit_y4 is not None and self.revenue_y4:
#             return self.reported_net_profit_y4 / self.revenue_y4
#
#     def _compute_net_profit_margin_y5(self) -> float:
#         if self.reported_net_profit_y5 is not None and self.revenue_y5:
#             return self.reported_net_profit_y5 / self.revenue_y5
#
#     def _compute_ebitda_margin_y1(self) -> float:
#         if self.ebitda_y1 is not None and self.revenue_y1:
#             return self.ebitda_y1 / self.revenue_y1
#
#     def _compute_ebitda_margin_y2(self) -> float:
#         if self.ebitda_y2 is not None and self.revenue_y2:
#             return self.ebitda_y2 / self.revenue_y2
#
#     def _compute_ebitda_margin_y3(self) -> float:
#         if self.ebitda_y3 is not None and self.revenue_y3:
#             return self.ebitda_y3 / self.revenue_y3
#
#     def _compute_ebitda_margin_y4(self) -> float:
#         if self.ebitda_y4 is not None and self.revenue_y4:
#             return self.ebitda_y4 / self.revenue_y4
#
#     def _compute_ebitda_margin_y5(self) -> float:
#         if self.ebitda_y5 is not None and self.revenue_y5:
#             return self.ebitda_y5 / self.revenue_y5
#
#     def _compute_ebit_margin_y1(self) -> float:
#         if self.ebit_y1 is not None and self.revenue_y1:
#             return self.ebit_y1 / self.revenue_y1
#
#     def _compute_ebit_margin_y2(self) -> float:
#         if self.ebit_y2 is not None and self.revenue_y2:
#             return self.ebit_y2 / self.revenue_y2
#
#     def _compute_ebit_margin_y3(self) -> float:
#         if self.ebit_y3 is not None and self.revenue_y3:
#             return self.ebit_y3 / self.revenue_y3
#
#     def _compute_ebit_margin_y4(self) -> float:
#         if self.ebit_y4 is not None and self.revenue_y4:
#             return self.ebit_y4 / self.revenue_y4
#
#     def _compute_ebit_margin_y5(self) -> float:
#         if self.ebit_y5 is not None and self.revenue_y5:
#             return self.ebit_y5 / self.revenue_y5
#
#     def _compute_eps_y1(self):
#         if self.reported_net_profit_y1 is not None and self.shares:
#             return self.reported_net_profit_y1 / self.shares
#
#     def _compute_eps_y2(self):
#         if self.reported_net_profit_y2 is not None and self.shares:
#             return self.reported_net_profit_y2 / self.shares
#
#     def _compute_eps_y3(self):
#         if self.reported_net_profit_y3 is not None and self.shares:
#             return self.reported_net_profit_y3 / self.shares
#
#     def _compute_eps_y4(self):
#         if self.reported_net_profit_y4 is not None and self.shares:
#             return self.reported_net_profit_y4 / self.shares
#
#     def _compute_eps_y5(self):
#         if self.reported_net_profit_y5 is not None and self.shares:
#             return self.reported_net_profit_y5 / self.shares
#
#     def _compute_eps_growth_y1(self) -> float:
#         if (fundamental := self.yearly_fundamental) and self.eps_y1 is not None and fundamental.eps:
#             return self.eps_y1 / fundamental.eps - 1
#
#     def _compute_eps_growth_y2(self) -> float:
#         if self.eps_y2 is not None and self.eps_y1:
#             return self.eps_y2 / self.eps_y1 - 1
#
#     def _compute_eps_growth_y3(self) -> float:
#         if self.eps_y3 is not None and self.eps_y2:
#             return self.eps_y3 / self.eps_y2 - 1
#
#     def _compute_eps_growth_y4(self) -> float:
#         if self.eps_y4 is not None and self.eps_y3:
#             return self.eps_y4 / self.eps_y3 - 1
#
#     def _compute_eps_growth_y5(self) -> float:
#         if self.eps_y5 is not None and self.eps_y4:
#             return self.eps_y5 / self.eps_y4 - 1
#
#     def _compute_net_debt_to_ebitda_ratio_y1(self) -> float:
#         if self.net_debt_y1 is not None and self.ebitda_y1:
#             return self.net_debt_y1 / self.ebitda_y1
#
#     def _compute_net_debt_to_ebitda_ratio_y2(self) -> float:
#         if self.net_debt_y2 is not None and self.ebitda_y2:
#             return self.net_debt_y2 / self.ebitda_y2
#
#     def _compute_net_debt_to_ebitda_ratio_y3(self) -> float:
#         if self.net_debt_y3 is not None and self.ebitda_y3:
#             return self.net_debt_y3 / self.ebitda_y3
#
#     def _compute_net_debt_to_ebitda_ratio_y4(self) -> float:
#         if self.net_debt_y4 is not None and self.ebitda_y4:
#             return self.net_debt_y4 / self.ebitda_y4
#
#     def _compute_net_debt_to_ebitda_ratio_y5(self) -> float:
#         if self.net_debt_y5 is not None and self.ebitda_y5:
#             return self.net_debt_y5 / self.ebitda_y5
#
#     def _compute_free_cash_flow_growth_y1(self) -> float:
#         if (
#             (daily_fundamental := self.daily_fundamental)
#             and self.free_cash_flow_y1 is not None
#             and daily_fundamental.free_cash_flow
#         ):
#             return self.free_cash_flow_y1 / daily_fundamental.free_cash_flow - 1
#
#     def _compute_free_cash_flow_growth_y2(self) -> float:
#         if self.free_cash_flow_y2 is not None and self.free_cash_flow_y1:
#             return self.free_cash_flow_y2 / self.free_cash_flow_y1 - 1
#
#     def _compute_free_cash_flow_growth_y3(self) -> float:
#         if self.free_cash_flow_y3 is not None and self.free_cash_flow_y2:
#             return self.free_cash_flow_y3 / self.free_cash_flow_y2 - 1
#
#     def _compute_free_cash_flow_growth_y4(self) -> float:
#         if self.free_cash_flow_y4 is not None and self.free_cash_flow_y3:
#             return self.free_cash_flow_y4 / self.free_cash_flow_y3 - 1
#
#     def _compute_free_cash_flow_growth_y5(self) -> float:
#         if self.free_cash_flow_y5 is not None and self.free_cash_flow_y4:
#             return self.free_cash_flow_y5 / self.free_cash_flow_y4 - 1
#
#     def _compute_free_cash_flow_to_sales_ratio_y1(self) -> float:
#         if self.free_cash_flow_y1 is not None and self.revenue_y1:
#             return self.free_cash_flow_y1 / self.revenue_y1
#
#     def _compute_free_cash_flow_to_sales_ratio_y2(self) -> float:
#         if self.free_cash_flow_y2 is not None and self.revenue_y2:
#             return self.free_cash_flow_y2 / self.revenue_y2
#
#     def _compute_free_cash_flow_to_sales_ratio_y3(self) -> float:
#         if self.free_cash_flow_y3 is not None and self.revenue_y3:
#             return self.free_cash_flow_y3 / self.revenue_y3
#
#     def _compute_free_cash_flow_to_sales_ratio_y4(self) -> float:
#         if self.free_cash_flow_y4 is not None and self.revenue_y4:
#             return self.free_cash_flow_y4 / self.revenue_y4
#
#     def _compute_free_cash_flow_to_sales_ratio_y5(self) -> float:
#         if self.free_cash_flow_y5 is not None and self.revenue_y5:
#             return self.free_cash_flow_y5 / self.revenue_y5
#
#     def _compute_capex_to_sales_y1(self) -> float:
#         if self.capital_expenditures_y1 is not None and self.revenue_y1:
#             return self.capital_expenditures_y1 / self.revenue_y1
#
#     def _compute_capex_to_sales_y2(self) -> float:
#         if self.capital_expenditures_y2 is not None and self.revenue_y2:
#             return self.capital_expenditures_y2 / self.revenue_y2
#
#     def _compute_capex_to_sales_y3(self) -> float:
#         if self.capital_expenditures_y3 is not None and self.revenue_y3:
#             return self.capital_expenditures_y3 / self.revenue_y3
#
#     def _compute_capex_to_sales_y4(self) -> float:
#         if self.capital_expenditures_y4 is not None and self.revenue_y4:
#             return self.capital_expenditures_y4 / self.revenue_y4
#
#     def _compute_capex_to_sales_y5(self) -> float:
#         if self.capital_expenditures_y5 is not None and self.revenue_y5:
#             return self.capital_expenditures_y5 / self.revenue_y5
#
#     def __compute_return_on_equity(self, year: int) -> float:
#         if (
#             (net_profit := getattr(self, f"reported_net_profit_y{year}")) is not None
#             and (previous := self.yearly_fundamental)
#             and previous.net_profit is not None
#             and previous.shareholder_equity
#         ):
#             return net_profit / previous.shareholder_equity
#
#     def _compute_return_on_equity_y1(self):
#         return self.__compute_return_on_equity(1)
#
#     def _compute_return_on_equity_y2(self):
#         return self.__compute_return_on_equity(2)
#
#     def _compute_return_on_equity_y3(self):
#         return self.__compute_return_on_equity(3)
#
#     def _compute_return_on_equity_y4(self):
#         return self.__compute_return_on_equity(4)
#
#     def _compute_return_on_equity_y5(self):
#         return self.__compute_return_on_equity(5)
#
#     def __compute_return_on_assets(self, year: int) -> float:
#         if (
#             (net_profit := getattr(self, f"reported_net_profit_y{year}")) is not None
#             and (previous := self.yearly_fundamental)
#             and previous.net_profit is not None
#             and previous.total_assets
#         ):
#             return net_profit / previous.total_assets
#
#     def _compute_return_on_assets_y1(self):
#         return self.__compute_return_on_assets(1)
#
#     def _compute_return_on_assets_y2(self):
#         return self.__compute_return_on_assets(2)
#
#     def _compute_return_on_assets_y3(self):
#         return self.__compute_return_on_assets(3)
#
#     def _compute_return_on_assets_y4(self):
#         return self.__compute_return_on_assets(4)
#
#     def _compute_return_on_assets_y5(self):
#         return self.__compute_return_on_assets(5)
#
#     def __compute_return_on_capital_employed(self, year: int) -> float:
#         if (
#             (previous := self.yearly_fundamental)
#             and (ebit := getattr(self, f"ebit_y{year}")) is not None
#             and previous.total_assets
#             and previous.current_liabilities
#             and (diff := previous.total_assets - previous.current_liabilities)
#         ):
#             return ebit / diff
#
#     def _compute_return_on_capital_employed_y1(self):
#         return self.__compute_return_on_capital_employed(1)
#
#     def _compute_return_on_capital_employed_y2(self):
#         return self.__compute_return_on_capital_employed(2)
#
#     def _compute_return_on_capital_employed_y3(self):
#         return self.__compute_return_on_capital_employed(3)
#
#     def _compute_return_on_capital_employed_y4(self):
#         return self.__compute_return_on_capital_employed(4)
#
#     def _compute_return_on_capital_employed_y5(self):
#         return self.__compute_return_on_capital_employed(5)
#
#     def __compute_return_on_invested_capital(self, year: int) -> float:
#         if (
#             (previous := self.yearly_fundamental)
#             and (ebit := getattr(self, f"ebit_y{year}")) is not None
#             and previous.company_tax_rate is not None
#             and previous.shareholder_equity is not None
#             and previous.total_liabilities is not None
#             and previous.cash_and_cash_equivalents is not None
#         ):
#             divisor = previous.shareholder_equity + previous.total_liabilities - previous.cash_and_cash_equivalents
#             if divisor:
#                 return ebit * (1 - previous.company_tax_rate) / divisor
#
#     def _compute_return_on_invested_capital_y1(self):
#         return self.__compute_return_on_invested_capital(1)
#
#     def _compute_return_on_invested_capital_y2(self):
#         return self.__compute_return_on_invested_capital(2)
#
#     def _compute_return_on_invested_capital_y3(self):
#         return self.__compute_return_on_invested_capital(3)
#
#     def _compute_return_on_invested_capital_y4(self):
#         return self.__compute_return_on_invested_capital(4)
#
#     def _compute_return_on_invested_capital_y5(self):
#         return self.__compute_return_on_invested_capital(5)
#
#     def __compute_interest_coverage_ratio(self, year: int) -> float:
#         if (
#             (previous := self.yearly_fundamental)
#             and (ebit := getattr(self, f"ebit_y{year}")) is not None
#             and previous.interest_expense
#         ):
#             return ebit / previous.interest_expense
#
#     def _compute_interest_coverage_ratio_y1(self):
#         return self.__compute_interest_coverage_ratio(1)
#
#     def _compute_interest_coverage_ratio_y2(self):
#         return self.__compute_interest_coverage_ratio(2)
#
#     def _compute_interest_coverage_ratio_y3(self):
#         return self.__compute_interest_coverage_ratio(3)
#
#     def _compute_interest_coverage_ratio_y4(self):
#         return self.__compute_interest_coverage_ratio(4)
#
#     def _compute_interest_coverage_ratio_y5(self):
#         return self.__compute_interest_coverage_ratio(5)
#
#
# class DailyFundamentalComputedMixin:
#     def _compute_free_cash_flow_ttm_growth(self) -> float:
#         if (
#             self.free_cash_flow is not None
#             and (last_ttl_daily_fundamental := self.last_ttl_daily_fundamental)
#             and (last_free_cash_flow := last_ttl_daily_fundamental.free_cash_flow)
#         ):
#             return self.free_cash_flow / last_free_cash_flow - 1
#
#     def _compute_free_cash_flow(self) -> float:
#         if (
#             self.free_cash is not None
#             and (price := self.price)
#             and (outstanding_shares_consolidated := price.outstanding_shares_consolidated)
#         ):
#             return self.free_cash / float(outstanding_shares_consolidated)
#
#     def _compute_revenue_growth_3y_cagr(self) -> float:
#         if (
#             (forecast := self.forecast)
#             and (fundamental := self.yearly_fundamental)
#             and (revenue_y3 := forecast.revenue_y3) is not None
#             and fundamental.revenue
#         ):
#             if (den := revenue_y3 / fundamental.revenue) >= 0:
#                 return math.pow(den, 1 / 3) - 1
#
#     def _compute_eps_3y_cagr(self) -> float:
#         if (
#             (forecast := self.forecast)
#             and (fundamental := self.yearly_fundamental)
#             and (eps_3y := forecast.eps_y3) is not None
#             and fundamental.eps
#         ):
#             if (den := eps_3y / fundamental.eps) >= 0:
#                 return math.pow(den, 1 / 3) - 1
#
#     def _compute_free_cash_flow_3y_cagr(self) -> float:
#         if (
#             (forecast := self.forecast)
#             and (fundamental := self.yearly_fundamental)
#             and (fcf_3y := forecast.free_cash_flow_y3) is not None
#             and fundamental.free_cash_flow
#         ):
#             if (den := fcf_3y / fundamental.free_cash_flow) >= 0:
#                 return math.pow(den, 1 / 3) - 1
#
#     def _compute_eps_ttm(self):
#         pass
#
#
class InstrumentPriceComputedMixin:
    def _compute_outstanding_shares(self):
        if self.outstanding_shares is None and (previous_price := self.previous_price):
            return previous_price.outstanding_shares

    def _compute_outstanding_shares_consolidated(self):
        if self.outstanding_shares_consolidated is None and self.outstanding_shares is not None:
            return self.outstanding_shares

    def _compute_gross_value(self):
        if self.net_value is not None and self.gross_value is None:
            return self.net_value

    def _compute_volume_50d(self):
        volumes = list(
            filter(
                None,
                self.instrument.valuations.filter(date__lte=self.date)
                .order_by("-date")
                .values_list("volume", flat=True)[:50],
            )
        )
        if len(volumes) > 0:
            return sum(volumes) / len(volumes)
        return 0.0

    def _compute_volume_200d(self):
        volumes = list(
            filter(
                None,
                self.instrument.valuations.filter(date__lte=self.date)
                .order_by("-date")
                .values_list("volume", flat=True)[:200],
            )
        )
        if len(volumes) > 0:
            return sum(volumes) / len(volumes)
        return 0.0

    def _compute_performance_1d(self):
        try:
            last_price = self.instrument.prices.get(
                calculated=self.calculated, date=(self.date - pd.tseries.offsets.BDay(1)).date()
            )
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass

    def _compute_performance_7d(self):
        try:
            last_price = self.instrument.prices.get(
                calculated=self.calculated, date=(self.date - pd.tseries.offsets.BDay(7)).date()
            )
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass

    def _compute_performance_30d(self):
        try:
            last_price = self.instrument.prices.get(
                calculated=self.calculated, date=(self.date - pd.tseries.offsets.BDay(30)).date()
            )
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass

    def _compute_performance_90d(self):
        try:
            last_price = self.instrument.prices.get(
                calculated=self.calculated, date=(self.date - pd.tseries.offsets.BDay(90)).date()
            )
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass

    def _compute_performance_365d(self):
        try:
            last_price = self.instrument.prices.get(
                calculated=self.calculated, date=(self.date - pd.tseries.offsets.BDay(365)).date()
            )
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass

    def _compute_performance_ytd(self):
        try:
            last_price = self.instrument.prices.get(
                calculated=self.calculated, date=(date(self.date.year, 1, 1) - pd.tseries.offsets.BDay(0)).date()
            )
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass

    def _compute_performance_inception(self):
        try:
            last_price = self.instrument.prices.get(calculated=self.calculated, date=self.instrument.inception_date)
            if self.net_value is not None and last_price.net_value:
                return self.net_value / last_price.net_value - 1
        except self.DoesNotExist:
            pass



from contextlib import suppress
from datetime import date
from enum import Enum
from functools import reduce
from itertools import batched
from typing import TYPE_CHECKING, Iterator

import pypika as pk
from django.db import ProgrammingError, connections
from pypika import Case, Column, MSSQLQuery
from pypika import functions as fn
from pypika.enums import Order, SqlTypes
from pypika.terms import LiteralValue, ValueWrapper
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.contrib.qa.dataloaders.utils import create_table
from wbfdm.dataloaders.protocols import MarketDataProtocol
from wbfdm.dataloaders.types import MarketDataDict
from wbfdm.enums import Frequency, MarketData

if TYPE_CHECKING:
    pass


class DS2MarketData(Enum):
    OPEN = "Open_"
    CLOSE = "Close_"
    HIGH = "High"
    LOW = "Low"
    BID = "Bid"
    ASK = "Ask"
    VWAP = "VWAP"
    VOLUME = "Volume"
    MARKET_CAPITALIZATION = "MktCap"
    SHARES_OUTSTANDING = "NumShrs"


class DatastreamMarketDataDataloader(MarketDataProtocol, Dataloader):
    def market_data(
        self,
        values: list[MarketData] = [MarketData.CLOSE],
        from_date: date | None = None,
        to_date: date | None = None,
        exact_date: date | None = None,
        frequency: Frequency = Frequency.DAILY,
        target_currency: str | None = None,
        **kwargs,
    ) -> Iterator[MarketDataDict]:
        """Get market data for instruments.

        Args:
            queryset (QuerySet["Instrument"]): The queryset of instruments.
            values (list[MarketData]): List of values to include in the results.
            from_date (date | None): The starting date for filtering prices. Defaults to None.
            to_date (date | None): The ending date for filtering prices. Defaults to None.
            frequency (Frequency): The frequency of the requested data

        Returns:
            Iterator[MarketDataDict]: An iterator of dictionaries conforming to the DailyValuationDict.
        """

        lookup = {
            f"{k[0]},{k[1]}": v for k, v in self.entities.values_list("dl_parameters__market_data__parameters", "id")
        }

        # Define tables
        pricing = pk.Table("vw_DS2Pricing")
        market_val = pk.Table("vw_DS2MktCap")

        mapping, create_mapping_table = create_table(
            "#ds2infoexchcode", Column("InfoCode", SqlTypes.INTEGER), Column("ExchIntCode", SqlTypes.INTEGER)
        )

        # Base query to get data we always need unconditionally
        query = (
            pk.MSSQLQuery.from_(pricing)
            .select(
                fn.Concat(pricing.InfoCode, ",", pricing.ExchIntCode).as_("external_identifier"),
                fn.Concat(
                    pricing.InfoCode, ",", pricing.ExchIntCode, "_", fn.Cast(pricing.MarketDate, SqlTypes.DATE)
                ).as_("id"),
                fn.Cast(pricing.MarketDate, SqlTypes.DATE).as_("valuation_date"),
                ValueWrapper("qa-ds2").as_("source"),
            )
            # We join on _codes, which removes all instruments not in _codes - implicit where
            .join(mapping)
            .on((pricing.InfoCode == mapping.InfoCode) & (pricing.ExchIntCode == mapping.ExchIntCode))
            .where(pricing.AdjType == 2)
            .orderby(pricing.MarketDate, order=Order.desc)
        )

        # if a target currency is required, we join on the fx tables and set the currency to the desired one
        # otherwise we just set the currency to whatever the currency is from the instrument
        fx_rate = None
        if target_currency:
            query = query.select(ValueWrapper(target_currency).as_("currency"))
            fx_code = pk.Table("DS2FxCode")
            fx_rate = pk.Table("DS2FxRate")
            query = (
                query
                # Join FX code table matching currencies and ensuring SPOT rate type
                .left_join(fx_code)
                .on(
                    (fx_code.FromCurrCode == pricing.Currency)
                    & (fx_code.ToCurrCode == target_currency)
                    & (fx_code.RateTypeCode == "SPOT")
                )
                # Join FX rate table matching internal code and date
                .left_join(fx_rate)
                .on((fx_rate.ExRateIntCode == fx_code.ExRateIntCode) & (fx_rate.ExRateDate == pricing.MarketDate))
                # We filter out rows which do not have a proper fx rate (we exclude same currency conversions)
                .where((Case().when(pricing.Currency == target_currency, 1).else_(fx_rate.midrate).isnotnull()))
            )

        else:
            query = query.select(pricing.Currency.as_("currency"))

        # if market cap or shares outstanding are required we need to join with an additional table
        if MarketData.MARKET_CAPITALIZATION in values or MarketData.SHARES_OUTSTANDING in values:
            query = query.join(market_val).on(
                (pricing.InfoCode == market_val.InfoCode) & (pricing.MarketDate == market_val.MarketDate)
            )

        for market_data in values:
            ds2_value = DS2MarketData[market_data.name].value
            value = LiteralValue(ds2_value)
            if fx_rate and market_data is not MarketData.SHARES_OUTSTANDING:
                value /= Case().when(pricing.Currency == target_currency, 1).else_(fx_rate.midrate)

            query = query.select(value.as_(market_data.value))

        # Add conditional where clauses
        if from_date:
            query = query.where(pricing.MarketDate >= from_date)

        if to_date:
            query = query.where(pricing.MarketDate <= to_date)

        if exact_date:
            query = query.where(pricing.MarketDate == exact_date)

        with connections["qa"].cursor() as cursor:
            with suppress(ProgrammingError):
                cursor.execute(create_mapping_table.get_sql())
                for batch in batched(
                    self.entities.values_list("dl_parameters__market_data__parameters", flat=True), 1000
                ):
                    cursor.execute(reduce(lambda x, y: x.insert(y), batch, MSSQLQuery.into(mapping)).get_sql())

            cursor.execute(query.get_sql())

            for row in dictfetchall(cursor, MarketDataDict):
                row["instrument_id"] = lookup[row["external_identifier"]]
                yield row

            cursor.execute(MSSQLQuery.drop_table(mapping).get_sql())



from datetime import date
from typing import Iterator

from wbcore.contrib.dataloader.dataloaders import DataloaderProxy

from wbfdm.dataloaders.protocols import (
    AdjustmentsProtocol,
    CorporateActionsProtocol,
    ESGControversyProtocol,
    ESGProtocol,
    FinancialsProtocol,
    MarketDataProtocol,
    OfficersProtocol,
    ReportDateProtocol,
    StatementsProtocol,
)
from wbfdm.dataloaders.types import (
    AdjustmentDataDict,
    CorporateActionDataDict,
    ESGControversyDataDict,
    ESGDataDict,
    FinancialDataDict,
    MarketDataDict,
    OfficerDataDict,
    ReportDateDataDict,
    StatementDataDict,
)
from wbfdm.enums import (
    ESG,
    CalendarType,
    DataType,
    EstimateType,
    Financial,
    Frequency,
    MarketData,
    PeriodType,
    SeriesType,
    StatementType,
)

from .cache import Cache


def _market_data_row_parser(row):
    if row.get("close") is None and (bid := row.get("bid")) is not None and (ask := row.get("ask")) is not None:
        price = (bid + ask) / 2
        row["close"] = price
        row["open"] = price
        row["low"] = price
        row["high"] = price
    return row


class InstrumentDataloaderProxy(
    DataloaderProxy[
        AdjustmentsProtocol
        | MarketDataProtocol
        | CorporateActionsProtocol
        | OfficersProtocol
        | StatementsProtocol
        | FinancialsProtocol
        | ReportDateProtocol
        | ESGControversyProtocol
        | ESGProtocol
    ]
):
    def reporting_dates(self, only_next: bool = True) -> Iterator[ReportDateDataDict]:
        for dl in self.iterate_dataloaders("reporting_dates"):
            yield from dl.reporting_dates(only_next=only_next)

    def adjustments(self, from_date: date | None = None, to_date: date | None = None) -> Iterator[AdjustmentDataDict]:
        for dl in self.iterate_dataloaders("adjustments"):
            yield from dl.adjustments(from_date=from_date, to_date=to_date)

    def market_data(
        self,
        values: list[MarketData] | None = None,
        from_date: date | None = None,
        to_date: date | None = None,
        exact_date: date | None = None,
        frequency: Frequency = Frequency.DAILY,
        target_currency: str | None = None,
    ) -> Iterator[MarketDataDict]:
        if not values:
            values = list(MarketData)
        for dl in self.iterate_dataloaders("market_data"):
            yield from map(
                lambda row: _market_data_row_parser(row),
                dl.market_data(
                    values=values,
                    from_date=from_date,
                    to_date=to_date,
                    exact_date=exact_date,
                    frequency=frequency,
                    target_currency=target_currency,
                ),
            )

    def corporate_actions(
        self, from_date: date | None = None, to_date: date | None = None
    ) -> Iterator[CorporateActionDataDict]:
        for dl in self.iterate_dataloaders("corporate_actions"):
            yield from dl.corporate_actions(from_date=from_date, to_date=to_date)

    def officers(self) -> Iterator[OfficerDataDict]:
        for dl in self.iterate_dataloaders("officers"):
            yield from dl.officers()

    def statements(
        self,
        statement_type: StatementType | None = None,
        from_date: date | None = None,
        to_date: date | None = None,
        from_year: int | None = None,
        to_year: int | None = None,
        period_type: PeriodType = PeriodType.ALL,
        data_type: DataType = DataType.STANDARDIZED,
        financials: list[Financial] | None = None,
        target_currency: str | None = None,
    ) -> Iterator[StatementDataDict]:
        for dl in self.iterate_dataloaders("statements"):
            yield from dl.statements(
                statement_type=statement_type,
                from_date=from_date,
                to_date=to_date,
                from_year=from_year,
                to_year=to_year,
                period_type=period_type,
                data_type=data_type,
                financials=financials,
                target_currency=target_currency,
            )

    def financials(
        self,
        values: list[Financial],
        from_date: date | None = None,
        to_date: date | None = None,
        from_year: int | None = None,
        to_year: int | None = None,
        from_index: int | None = None,
        to_index: int | None = None,
        from_valid: date | None = None,
        to_valid: date | None = None,
        period_type: PeriodType = PeriodType.ANNUAL,
        calendar_type: CalendarType = CalendarType.FISCAL,
        series_type: SeriesType = SeriesType.COMPLETE,
        data_type: DataType = DataType.STANDARDIZED,
        estimate_type: EstimateType = EstimateType.VALID,
        target_currency: str | None = None,
    ) -> Iterator[FinancialDataDict]:
        for dl in self.iterate_dataloaders("financials"):
            yield from dl.financials(
                values=values,
                from_date=from_date,
                to_date=to_date,
                from_year=from_year,
                to_year=to_year,
                from_index=from_index,
                to_index=to_index,
                from_valid=from_valid,
                to_valid=to_valid,
                period_type=period_type,
                calendar_type=calendar_type,
                series_type=series_type,
                data_type=data_type,
                estimate_type=estimate_type,
                target_currency=target_currency,
            )

    def esg_controversies(self) -> Iterator[ESGControversyDataDict]:
        for dl in self.iterate_dataloaders("esg_controversies"):
            yield from dl.esg_controversies()

    def esg(
        self,
        values: list[ESG],
    ) -> Iterator[ESGDataDict]:
        for dl in self.iterate_dataloaders("esg"):
            if (
                (cache_identifier_key := getattr(dl, "CACHE_IDENTIFIER_KEY", None))
                and (cache_symbol_key := getattr(dl, "CACHE_SYMBOL_KEY", None))
                and (cache_value_key := getattr(dl, "CACHE_VALUE_KEY", None))
            ):
                cache = Cache(
                    identifier_key=cache_identifier_key,
                    symbol_key=cache_symbol_key,
                    value_key=cache_value_key,
                    timeout=getattr(dl, "CACHE_TIMEOUT", 10 * 24 * 3600),  # default to 10days
                )
                cache.initialize(dl.entity_ids, [v.value for v in values])

                yield from cache.fetch_from_cache()
                if len(cache.missing_symbols) > 0 and len(cache.missing_ids) > 0:
                    dl.entities = dl.entities.filter(id__in=cache.missing_ids)
                    yield from map(
                        lambda row: cache.write(row),
                        dl.esg(values=[v for v in values if v.value in cache.missing_symbols]),
                    )
                cache.close()
            else:
                yield from dl.esg(values=values)



from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from wbcore.models import WBModel


class ExchangeManager(models.Manager):
    def get_by_bbg(self, bbg):
        """
        If passed in a bbg ticker, we don't know whether it is a normal exchange code or a composite code.
        Therefore we return either the exchange where this code is the bbg_exchange_codes, or the exchange where this composite code
        exists, but only if it is the primary exchange
        """
        return self.get(Q(bbg_exchange_codes__contains=[bbg]) | (Q(bbg_composite=bbg) & Q(bbg_composite_primary=True)))


class Exchange(WBModel):
    source_id = models.CharField(max_length=64, null=True, blank=True)
    source = models.CharField(max_length=64, null=True, blank=True)
    name = models.CharField(
        max_length=265, null=True, blank=True, verbose_name="Exchange Name", help_text="Name of the Exchange."
    )

    opening_time = models.TimeField(
        blank=True, null=True, verbose_name="Exchange Opening time", help_text="The opening time of the exchange"
    )
    closing_time = models.TimeField(
        blank=True, null=True, verbose_name="Exchange Closing time", help_text="The closing time of the exchange"
    )

    mic_code = models.CharField(
        max_length=4,
        null=True,
        blank=True,
        verbose_name="MIC (ISO)",
        unique=True,
        help_text="Market Identifier Code.",
    )
    mic_name = models.CharField(
        max_length=126,
        null=True,
        blank=True,
        verbose_name="MIC (ISO) Name",
        help_text="Market Identifier Name.",
    )
    operating_mic_code = models.CharField(
        max_length=4,
        null=True,
        blank=True,
        verbose_name="Operating MIC Code",
        help_text="Operating Market Identifier Code.",
    )
    operating_mic_name = models.CharField(
        max_length=126,
        null=True,
        blank=True,
        verbose_name="Operating MIC Name",
        help_text="Operating Market Identifier Name.",
    )
    ########################################################
    #                       Bloomberg                      #
    ########################################################

    bbg_exchange_codes = ArrayField(
        models.CharField(
            max_length=4,
        ),
        blank=True,
        default=list,
        verbose_name="BBG Exchange Code",
        help_text="Bloomberg Exchange Code.",
    )

    bbg_composite_primary = models.BooleanField(
        default=False,
        verbose_name="BBG Primary composite",
        help_text="Indicates the primary exchange for this BBG Composite Code.",
    )

    bbg_composite = models.CharField(
        max_length=4,
        null=True,
        blank=True,
        verbose_name="BBG Composite Code",
        help_text="Bloomberg Composite Code.",
    )

    ########################################################
    #                   Reuters/Refinitiv                  #
    ########################################################

    refinitiv_identifier_code = models.CharField(
        max_length=4,
        null=True,
        blank=True,
        verbose_name="RIC Exchange Code",
        help_text="Reuters Exchange Code.",
    )
    refinitiv_mnemonic = models.CharField(
        max_length=4,
        null=True,
        blank=True,
        verbose_name="Refinitiv Mnemonic",
        help_text="Reuters Exchange Mnemonic Code.",
    )
    ########################################################
    #                      INFORMATION                     #
    ########################################################

    country = models.ForeignKey(
        to="geography.Geography",
        null=True,
        blank=True,
        limit_choices_to={"level": 1},
        on_delete=models.SET_NULL,
    )

    city = models.ForeignKey(
        to="geography.Geography",
        max_length=255,
        null=True,
        blank=True,
        limit_choices_to={"level": 3},
        related_name="city_exchanges",
        on_delete=models.SET_NULL,
        verbose_name="City",
        help_text="The city where this Exchange is located at.",
    )

    city = models.ForeignKey(
        "geography.Geography",
        null=True,
        blank=True,
        related_name="exchanges",
        verbose_name="City",
        on_delete=models.PROTECT,
        help_text="The city where this Exchange is located at.",
        limit_choices_to={"level": 3},
    )

    website = models.URLField(
        null=True,
        blank=True,
        verbose_name="Website",
        help_text="The Website of the Exchange",
    )

    comments = models.TextField(
        default="",
        blank=True,
        verbose_name="Comments",
        help_text="Any comments for this exchange",
    )

    last_updated = models.DateTimeField(auto_now=True)

    objects = ExchangeManager()

    @property
    def identifier_repr(self) -> str:
        if self.bbg_exchange_codes:
            return self.bbg_exchange_codes[0]
        if self.mic_code:
            return self.mic_code
        if self.operating_mic_code:
            return self.operating_mic_code
        if self.bbg_composite:
            return self.bbg_composite
        if self.refinitiv_identifier_code:
            return self.refinitiv_identifier_code
        if self.refinitiv_mnemonic:
            return self.refinitiv_mnemonic

    def __str__(self) -> str:
        repr = self.identifier_repr
        if self.name:
            repr = f"{self.name} ({self.identifier_repr})"
        elif self.mic_name:
            repr = f"{self.mic_name} ({self.identifier_repr})"
        return repr

    class Meta:
        verbose_name = _("Exchange")
        verbose_name_plural = _("Exchanges")
        indexes = [models.Index(fields=["source_id", "source"])]
        constraints = [
            models.UniqueConstraint(fields=["source", "source_id"], name="unique_exchange_source"),
        ]

    @classmethod
    def dict_to_model(cls, exchange_data):
        if isinstance(exchange_data, int):
            return Exchange.objects.filter(id=exchange_data).first()
        elif (bbg_code := exchange_data.get("bbg_exchange_codes", None)) or (
            bbg_code := exchange_data.get("bbg_exchange_code", None)
        ):
            return Exchange.objects.filter(bbg_exchange_codes__contains=[bbg_code]).first()
        elif mic_code := exchange_data.pop("exchange__mic_code", None):
            return Exchange.objects.filter(mic_code=mic_code).first()
        elif bbg_composite := exchange_data.pop("exchange__bbg_composite", None):
            return Exchange.objects.filter(bbg_composite=bbg_composite).first()

    @classmethod
    def get_endpoint_basename(cls):
        return "wbfdm:exchange"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbfdm:exchangerepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}} ({{mic_code}})"

    @property
    def time_zone(self):
        if self.city:
            return self.city.time_zone



# Generated by Django 4.2.11 on 2024-04-05 07:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0016_instrumenttype_name_repr"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="instrument",
            index=models.Index(fields=["is_security"], name="instrument_security_index"),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class InstrumentPriceInstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbfdm:instrument-price-list", [self.view.kwargs["instrument_id"]], request=self.request)


class InstrumentPriceStatisticsInstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-pricestatisticchart-list",
            [self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class MonthlyPerformancesInstrumentEndpointConfig(InstrumentPriceInstrumentEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse("wbfdm:monthly_performances-list", [self.view.kwargs["instrument_id"]], request=self.request)


class FinancialStatisticsInstrumentEndpointConfig(InstrumentPriceInstrumentEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-financialstatistics-list",
            [self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class InstrumentPriceInstrumentDistributionReturnsChartEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-distributionreturnschart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class BestAndWorstReturnsInstrumentEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-bestandworstreturns-list",
            [self.view.kwargs["instrument_id"]],
            request=self.request,
        )
